/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef CALODET_DESUBCALORIMETER_H
#define CALODET_DESUBCALORIMETER_H
#include "CaloDet/CLIDDeSubCalorimeter.h"

#include "DetDesc/DetectorElement.h"
#include "Detector/Calo/CaloCellID.h"
#include "GaudiKernel/MsgStream.h"

class MsgStream;
class DeSubSubCalorimeter;

/** @class DeSubCalorimeter DeSubCalorimeter.h "CaloDet/DeSubCalorimeter.h"
 *
 *  specific detector element class (calo halves)
 *
 *  @author Olivier Deschamps odescham@in2p3.fr
 *
 */

class DeSubCalorimeter : public DetDesc::DetectorElementPlus {
public:
  enum class Side { C = -1, A = +1 };
  /// Constructors
  DeSubCalorimeter( const std::string& name = "" );
  /// object identification
  static const CLID& classID() { return CLID_DeSubCalorimeter; }
  /// printout to standard STL stream
  std::ostream& printOut( std::ostream& os = std::cerr ) const override;
  /// printout to standard Gaudi stream
  MsgStream& printOut( MsgStream& ) const override;
  /// object identification
  const CLID& clID() const override;
  /// standard initialization
  StatusCode initialize() override;
  // accessing the geometry parameters
  //----------------------------------
  /// get side
  Side        side() const { return m_side; };
  std::string sideName() const {
    switch ( m_side ) {
    case Side::A:
      return "A-side";
    case Side::C:
      return "C-side";
    default:
      return "??"; // dark side??
    }
  };
  /// set side (0=C=Right / 1=A=Left)
  void setSide( Side s ) { m_side = s; };
  // get constant access to subcalorimeters
  const auto& subSubCalos() const { return m_subSubCalos; };
  const auto& subSubCalosSorted() const { return m_subSubCalosSorted; };
  double      xSize() const { return m_xSize; };
  double      ySize() const { return m_ySize; };
  // geometry setters (temp)
  void setXSize( const double xSize ) { m_xSize = xSize; }
  void setYSize( const double ySize ) { m_ySize = ySize; }

  friend std::ostream& operator<<( std::ostream& os, const DeSubCalorimeter& de ) { return de.printOut( os ); }
  friend MsgStream&    operator<<( MsgStream& os, const DeSubCalorimeter& de ) { return de.printOut( os ); }

private:
  std::vector<const DeSubSubCalorimeter*> m_subSubCalosSorted;
  std::vector<const DeSubSubCalorimeter*> m_subSubCalos;
  Side                                    m_side  = Side::C; ///< Calo side (0=C-side=Right , 1=A-side=Left)
  double                                  m_xSize = std::numeric_limits<double>::signaling_NaN();
  double                                  m_ySize = std::numeric_limits<double>::signaling_NaN();
};
// ============================================================================
inline std::ostream& operator<<( std::ostream& os, const DeSubCalorimeter* de ) {
  return de ? ( os << *de ) : ( os << " DeSubCalorimeter* points to NULL!" << std::endl );
}
inline MsgStream& operator<<( MsgStream& os, const DeSubCalorimeter* de ) {
  return de ? ( os << *de ) : ( os << " DeSubCalorimeter* points to NULL!" << endmsg );
}
// ============================================================================
#endif //   CALODET_DESUBCALORIMETER_H
