/*****************************************************************************\
* (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "Event/GenCountersFSR.h"

namespace LHCb {

  const std::map<GenCountersFSR::CounterKey, std::string> GenCountersFSR::ck2s{

      {CounterKey::AllEvt, "all events (including empty events)"},
      {CounterKey::ZeroInt, "events with 0 interactions"},

      {CounterKey::EvtGenerated, "generated events"},
      {CounterKey::IntGenerated, "generated interactions"},
      {CounterKey::EvtAccepted, "accepted events"},
      {CounterKey::IntAccepted, "interactions in accepted events"},

      {CounterKey::BeforeFullEvt, "before full event generator level cut"},
      {CounterKey::AfterFullEvt, "full event generator level cut"},

      {CounterKey::OnebGen, "generated interactions with >= 1b"},
      {CounterKey::ThreebGen, "generated interactions with >= 3b"},
      {CounterKey::PromptBGen, "generated interactions with 1 prompt B"},
      {CounterKey::OnecGen, "generated interactions with >= 1c"},
      {CounterKey::ThreecGen, "generated interactions with >= 3c"},
      {CounterKey::PromptCGen, "generated interactions with 1 prompt C"},
      {CounterKey::bAndcGen, "generated interactions with b and c"},
      {CounterKey::OnebAcc, "accepted interactions with >= 1b"},
      {CounterKey::ThreebAcc, "accepted interactions with >= 3b"},
      {CounterKey::PromptBAcc, "accepted interactions with 1 prompt B"},
      {CounterKey::OnecAcc, "accepted  interactions with >= 1c"},
      {CounterKey::ThreecAcc, "accepted interactions with >= 3c"},
      {CounterKey::PromptCAcc, "accepted interactions with 1 prompt C"},
      {CounterKey::bAndcAcc, "accepted interactions with b and c"},

      {CounterKey::BeforeLevelCut, "generated inclusive events"},
      {CounterKey::AfterLevelCut, "generator level cut"},
      {CounterKey::EvtInverted, "z-inverted events"},

      {CounterKey::B0Gen, "generated B0"},
      {CounterKey::antiB0Gen, "generated anti-B0"},
      {CounterKey::BplusGen, "generated B+"},
      {CounterKey::BminusGen, "generated B-"},
      {CounterKey::Bs0Gen, "generated Bs0"},
      {CounterKey::antiBs0Gen, "generated anti-Bs0"},
      {CounterKey::bBaryonGen, "generated b-Baryon"},
      {CounterKey::antibBaryonGen, "generated anti-b-Baryon"},
      {CounterKey::BcplusGen, "generated Bc+"},
      {CounterKey::BcminusGen, "generated Bc-"},
      {CounterKey::bbGen, "generated (bb)"},
      {CounterKey::B0Acc, "accepted B0"},
      {CounterKey::antiB0Acc, "accepted anti-B0"},
      {CounterKey::BplusAcc, "accepted B+"},
      {CounterKey::BminusAcc, "accepted B-"},
      {CounterKey::Bs0Acc, "accepted Bs0"},
      {CounterKey::antiBs0Acc, "accepted anti-Bs0s"},
      {CounterKey::bBaryonAcc, "accepted b-Baryon"},
      {CounterKey::antibBaryonAcc, "accepted anti-b-Baryon"},
      {CounterKey::BcplusAcc, "accepted Bc+s"},
      {CounterKey::BcminusAcc, "accepted Bc-"},
      {CounterKey::bbAcc, "accepted (bb)"},

      {CounterKey::D0Gen, "generated D0"},
      {CounterKey::antiD0Gen, "generated anti-D0"},
      {CounterKey::DplusGen, "generated D+"},
      {CounterKey::DminusGen, "generated D-"},
      {CounterKey::DsplusGen, "generated Ds+"},
      {CounterKey::DsminusGen, "generated Ds-"},
      {CounterKey::cBaryonGen, "generated c-Baryon"},
      {CounterKey::anticBaryonGen, "generated anti-c-Baryon"},
      {CounterKey::ccGen, "generated (cc)"},
      {CounterKey::D0Acc, "accepted D0"},
      {CounterKey::antiD0Acc, "accepted anti-D0"},
      {CounterKey::DplusAcc, "accepted D+"},
      {CounterKey::DminusAcc, "acceptedD-"},
      {CounterKey::DsplusAcc, "accepted Ds+"},
      {CounterKey::DsminusAcc, "accepted Ds-"},
      {CounterKey::cBaryonAcc, "accepted c-Baryon"},
      {CounterKey::anticBaryonAcc, "accepted anti-c-Baryon"},
      {CounterKey::ccAcc, "accepted (cc)"},

      {CounterKey::BGen, "generated B(L=0, J=0)"},
      {CounterKey::BstarGen, "generated B* (L=0, J=1)"},
      {CounterKey::B2starGen, "generated B** (L=1, J=0,1,2)"},
      {CounterKey::BAcc, "accepted B(L=0, J=0)"},
      {CounterKey::BstarAcc, "accepted B* (L=0, J=1)"},
      {CounterKey::B2starAcc, "accepted B** (L=1, J=0,1,2)"},

      {CounterKey::DGen, "generated D(L=0, J=0)"},
      {CounterKey::DstarGen, "generated D* (L=0, J=1)"},
      {CounterKey::D2starGen, "generated D** (L=1, J=0,1,2)"},
      {CounterKey::DAcc, "accepted D(L=0, J=0)"},
      {CounterKey::DstarAcc, "accepted D* (L=0, J=1)"},
      {CounterKey::D2starAcc, "accepted D** (L=1, J=0,1,2)"},

      {CounterKey::EvtSignal, "generated signal events"},
      {CounterKey::EvtantiSignal, "generated anti-signal events"},

      {CounterKey::BeforePCut, "generated signal events"},
      {CounterKey::AfterPCut, "generator particle level cut"},
      {CounterKey::BeforeantiPCut, "generated anti-signal events"},
      {CounterKey::AfterantiPCut, "generator anti-particle level cut"},

      {CounterKey::AllsubProcess, "Total cross-section"}};

  const GaudiUtils::VectorMap<std::string, GenCountersFSR::CounterKey> GenCountersFSR::s2ck = {
      {"AllEvt", CounterKey::AllEvt},
      {"ZeroInt", CounterKey::ZeroInt},
      {"EvtGenerated", CounterKey::EvtGenerated},
      {"IntGenerated", CounterKey::IntGenerated},
      {"EvtAccepted", CounterKey::EvtAccepted},
      {"IntAccepted", CounterKey::IntAccepted},
      {"BeforeFullEvt", CounterKey::BeforeFullEvt},
      {"AfterFullEvt", CounterKey::AfterFullEvt},
      {"OnebGen", CounterKey::OnebGen},
      {"ThreebGen", CounterKey::ThreebGen},
      {"PromptBGen", CounterKey::PromptBGen},
      {"OnecGen", CounterKey::OnecGen},
      {"ThreecGen", CounterKey::ThreecGen},
      {"PromptCGen", CounterKey::PromptCGen},
      {"bAndcGen", CounterKey::bAndcGen},
      {"OnebAcc", CounterKey::OnebAcc},
      {"ThreebAcc", CounterKey::ThreebAcc},
      {"PromptBAcc", CounterKey::PromptBAcc},
      {"OnecAcc", CounterKey::OnecAcc},
      {"ThreecAcc", CounterKey::ThreecAcc},
      {"PromptCAcc", CounterKey::PromptCAcc},
      {"bAndcAcc", CounterKey::bAndcAcc},
      {"BeforeLevelCut", CounterKey::BeforeLevelCut},
      {"AfterLevelCut", CounterKey::AfterLevelCut},
      {"EvtInverted", CounterKey::EvtInverted},
      {"B0Gen", CounterKey::B0Gen},
      {"antiB0Gen", CounterKey::antiB0Gen},
      {"BplusGen", CounterKey::BplusGen},
      {"BminusGen", CounterKey::BminusGen},
      {"Bs0Gen", CounterKey::Bs0Gen},
      {"antiBs0Gen", CounterKey::antiBs0Gen},
      {"bBaryonGen", CounterKey::bBaryonGen},
      {"antibBaryonGen", CounterKey::antibBaryonGen},
      {"BcplusGen", CounterKey::BcplusGen},
      {"BcminusGen", CounterKey::BcminusGen},
      {"bbGen", CounterKey::bbGen},
      {"B0Acc", CounterKey::B0Acc},
      {"antiB0Acc", CounterKey::antiB0Acc},
      {"BplusAcc", CounterKey::BplusAcc},
      {"BminusAcc", CounterKey::BminusAcc},
      {"Bs0Acc", CounterKey::Bs0Acc},
      {"antiBs0Acc", CounterKey::antiBs0Acc},
      {"bBaryonAcc", CounterKey::bBaryonAcc},
      {"antibBaryonAcc", CounterKey::antibBaryonAcc},
      {"BcplusAcc", CounterKey::BcplusAcc},
      {"BcminusAcc", CounterKey::BcminusAcc},
      {"bbAcc", CounterKey::bbAcc},
      {"D0Gen", CounterKey::D0Gen},
      {"antiD0Gen", CounterKey::antiD0Gen},
      {"DplusGen", CounterKey::DplusGen},
      {"DminusGen", CounterKey::DminusGen},
      {"DsplusGen", CounterKey::DsplusGen},
      {"DsminusGen", CounterKey::DsminusGen},
      {"cBaryonGen", CounterKey::cBaryonGen},
      {"anticBaryonGen", CounterKey::anticBaryonGen},
      {"ccGen", CounterKey::ccGen},
      {"D0Acc", CounterKey::D0Acc},
      {"antiD0Acc", CounterKey::antiD0Acc},
      {"DplusAcc", CounterKey::DplusAcc},
      {"DminusAcc", CounterKey::DminusAcc},
      {"DsplusAcc", CounterKey::DsplusAcc},
      {"DsminusAcc", CounterKey::DsminusAcc},
      {"cBaryonAcc", CounterKey::cBaryonAcc},
      {"anticBaryonAcc", CounterKey::anticBaryonAcc},
      {"ccAcc", CounterKey::ccAcc},
      {"BGen", CounterKey::BGen},
      {"BstarGen", CounterKey::BstarGen},
      {"B2starGen", CounterKey::B2starGen},
      {"BAcc", CounterKey::BAcc},
      {"BstarAcc", CounterKey::BstarAcc},
      {"B2starAcc", CounterKey::B2starAcc},
      {"DGen", CounterKey::DGen},
      {"DstarGen", CounterKey::DstarGen},
      {"D2starGen", CounterKey::D2starGen},
      {"DAcc", CounterKey::DAcc},
      {"DstarAcc", CounterKey::DstarAcc},
      {"D2starAcc", CounterKey::D2starAcc},
      {"EvtSignal", CounterKey::EvtSignal},
      {"EvtantiSignal", CounterKey::EvtantiSignal},
      {"BeforePCut", CounterKey::BeforePCut},
      {"AfterPCut", CounterKey::AfterPCut},
      {"BeforeantiPCut", CounterKey::BeforeantiPCut},
      {"AfterantiPCut", CounterKey::AfterantiPCut},
      {"Unknown", CounterKey::Unknown},
      {"AllsubProcess", CounterKey::AllsubProcess}};

  std::ostream& operator<<( std::ostream& s, GenCountersFSR::CounterKey e ) {
    auto iter = std::find_if( GenCountersFSR::s2ck.begin(), GenCountersFSR::s2ck.end(),
                              [&]( const auto& i ) { return i.second == e; } );
    if ( iter == GenCountersFSR::s2ck.end() )
      return s << "ERROR wrong value " << int( e ) << " for enum LHCb::GenCountersFSR::CounterKey";
    return s << iter->first;
  }
} // namespace LHCb
