/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once
#include "GaudiKernel/KeyedContainer.h"
#include "GaudiKernel/KeyedObject.h"
#include <ostream>

namespace LHCb {

  // Forward declarations

  // Class ID definition
  static const CLID CLID_TimeSpanFSR = 13504;

  // Namespace for locations in TDS
  namespace TimeSpanFSRLocation {
    inline const std::string Default = "/FileRecords/TimeSpanFSR";
  }

  /** @class TimeSpanFSR TimeSpanFSR.h
   *
   * Class to store time span of an input file in FSR
   *
   * @author Jaap Panman
   *
   */

  class TimeSpanFSR final : public KeyedObject<int> {
  public:
    /// typedef for KeyedContainer of TimeSpanFSR
    typedef KeyedContainer<TimeSpanFSR, Containers::HashMap> Container;

    /// Default Constructor
    TimeSpanFSR() : m_earliest( 0 ), m_latest( 0 ) {}

    // Retrieve pointer to class definition structure
    const CLID&        clID() const override { return LHCb::TimeSpanFSR::classID(); }
    static const CLID& classID() { return CLID_TimeSpanFSR; }

    /// Assignment operator
    TimeSpanFSR& operator=( const LHCb::TimeSpanFSR& rhs ) {
      m_earliest = rhs.m_earliest;
      m_latest   = rhs.m_latest;
      return *this;
    }

    /// add a time and store the earliest/latest times
    TimeSpanFSR& addTime( unsigned long long present ) {
      m_earliest = ( m_earliest > 0 ? std::min( present, m_earliest ) : present );
      m_latest   = std::max( present, m_latest );
      return *this;
    }

    /// Addition operator
    TimeSpanFSR& operator+=( const unsigned long long rhs ) {
      addTime( rhs );
      return *this;
    }

    /// add two FSRs and store the earliest/latest times
    TimeSpanFSR& operator+=( const LHCb::TimeSpanFSR& rhs ) {
      if ( rhs.m_earliest > 0 ) addTime( rhs.m_earliest );
      addTime( rhs.m_latest );
      return *this;
    }

    /// readable printout
    std::ostream& fillStream( std::ostream& s ) const override {
      return s << "{ "
               << " earliest : " << m_earliest << " - latest : " << m_latest << " }";
    }

    /// Retrieve const  Earliest time found on a file
    [[nodiscard]] constexpr unsigned long long earliest() const { return m_earliest; }

    /// Update  Earliest time found on a file
    TimeSpanFSR& setEarliest( unsigned long long value ) {
      m_earliest = value;
      return *this;
    }

    /// Retrieve const  Latest time found on a file
    [[nodiscard]] constexpr unsigned long long latest() const { return m_latest; }

    /// Update  Latest time found on a file
    TimeSpanFSR& setLatest( unsigned long long value ) {
      m_latest = value;
      return *this;
    }

    friend std::ostream& operator<<( std::ostream& str, const TimeSpanFSR& obj ) { return obj.fillStream( str ); }

  private:
    unsigned long long m_earliest; ///< Earliest time found on a file
    unsigned long long m_latest;   ///< Latest time found on a file

  }; // class TimeSpanFSR

  /// Definition of Keyed Container for TimeSpanFSR
  typedef KeyedContainer<TimeSpanFSR, Containers::HashMap> TimeSpanFSRs;

} // namespace LHCb
