/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

// Include files
#include "GaudiKernel/KeyedContainer.h"
#include "GaudiKernel/KeyedObject.h"
#include "GaudiKernel/NamedRange.h"
#include "GaudiKernel/Point3DTypes.h"
#include "GaudiKernel/SymmetricMatrixTypes.h"
#include "GaudiKernel/SystemOfUnits.h"
#include "GaudiKernel/VectorMap.h"
#include "Kernel/RelationObjectTypeTraits.h"
#include "LHCbMath/MatVec.h"
#include <ostream>
#include <utility>
#include <vector>

// Forward declarations

namespace LHCb {

  // Forward declarations

  // Class ID definition
  static const CLID CLID_VertexBase = 811;

  /** @class VertexBase VertexBase.h
   *
   * Base class from which all vertices are derived
   *
   * @author Juan Palacios
   *
   */

  class VertexBase : public KeyedObject<int> {
  public:
    /// typedef for std::vector of VertexBase
    using Vector      = std::vector<VertexBase*>;
    using ConstVector = std::vector<const VertexBase*>;

    /// typedef for KeyedContainer of VertexBase
    typedef KeyedContainer<VertexBase, Containers::HashMap> Container;

    /// User information
    typedef GaudiUtils::VectorMap<int, double> ExtraInfo;
    /// For uniform access to containers in TES (KeyedContainer,SharedContainer)
    using Range = Gaudi::NamedRange_<ConstVector>;

    /// Describe how the vertex was made
    enum VertexType { Primary = 100 };

    /// Copy constructor
    VertexBase( const VertexBase& rhs )
        : KeyedObject<int>{}
        , m_position{rhs.m_position}
        , m_covMatrix{rhs.m_covMatrix}
        , m_chi2{rhs.m_chi2}
        , m_nDoF{rhs.m_nDoF}
        , m_extraInfo{rhs.m_extraInfo} {}

    /// create a vertex with a selected key
    explicit VertexBase( int key ) : KeyedObject<int>( key ) {}

    /// Constructor from a point
    VertexBase( Gaudi::XYZPoint point ) : m_position( std::move( point ) ) {}

    /// Default Constructor
    VertexBase() = default;

    // Retrieve pointer to class definition structure
    [[nodiscard]] const CLID& clID() const override { return LHCb::VertexBase::classID(); }
    static const CLID&        classID() { return CLID_VertexBase; }

    /// Clone vertex
    [[nodiscard]] virtual VertexBase* clone() const;

    /// Assignment operator
    VertexBase& operator=( const VertexBase& rhs );

    /// Retrieve the Chi^2/DoF of vertex
    [[nodiscard]] double chi2PerDoF() const;

    /// Set the Chi^2 and the DoF of the vertex (fit)
    void setChi2AndDoF( double chi2, int ndof );

    /// ExtraInformation. Don't use directly, use hasInfo, info, addInfo...
    [[nodiscard]] const ExtraInfo& extraInfo() const;

    /// has information for specified key
    [[nodiscard]] bool hasInfo( int key ) const;

    ///  Add new information associated with the specified key. This method cannot be used to modify information for a
    ///  pre-existing key.
    bool addInfo( int key, double info );

    /// extract the information associated with the given key. If there is no such infomration the default value will be
    /// returned.
    [[nodiscard]] double info( int key, double def ) const;

    /// erase the information associated with the given key
    unsigned long eraseInfo( int key );

    /// Is the vertex a primary?
    [[nodiscard]] virtual bool isPrimary() const;

    /// Print this Particle in a human readable way
    std::ostream& fillStream( std::ostream& s ) const override;

    /// Retrieve const  Position in LHCb reference system
    [[nodiscard]] const Gaudi::XYZPoint& position() const { return m_position; }

    /// Update  Position in LHCb reference system
    VertexBase& setPosition( const Gaudi::XYZPoint& value ) {
      m_position = value;
      return *this;
    }

    /// Retrieve const  Covariance matrix containing errors on vertex position
    [[nodiscard]] const Gaudi::SymMatrix3x3& covMatrix() const { return m_covMatrix; }

    /// Update  Covariance matrix containing errors on vertex position
    VertexBase& setCovMatrix( const Gaudi::SymMatrix3x3& value ) {
      m_covMatrix = value;
      return *this;
    }

    /// Retrieve const  Chi square of vertex fit
    [[nodiscard]] double chi2() const { return m_chi2; }

    /// Update  Chi square of vertex fit
    VertexBase& setChi2( double value ) {
      m_chi2 = value;
      return *this;
    }

    /// Retrieve const  Number of degree of freedom
    [[nodiscard]] int nDoF() const { return m_nDoF; }

    /// Update  Number of degree of freedom
    VertexBase& setNDoF( int value ) {
      m_nDoF = value;
      return *this;
    }

    /// Update  Some addtional user information. Don't use directly. Use *Info() methods.
    VertexBase& setExtraInfo( const ExtraInfo& value ) {
      m_extraInfo = value;
      return *this;
    }

    friend std::ostream& operator<<( std::ostream& str, const VertexBase& obj ) { return obj.fillStream( str ); }

    // Interface compatibility with ThOr
    [[nodiscard, gnu::always_inline]] friend LinAlg::Vec<SIMDWrapper::scalar::float_v, 3>
    endVertexPos( const VertexBase& vtx ) {
      auto pos = vtx.position();
      return {pos.x(), pos.y(), pos.z()};
    }
    [[nodiscard, gnu::always_inline]] friend auto posCovMatrix( const VertexBase& vtx ) {
      return LHCb::LinAlg::convert<SIMDWrapper::scalar::float_v>( vtx.covMatrix() );
    }

  private:
    Gaudi::XYZPoint     m_position{0.0, 0.0, -100 * Gaudi::Units::m}; ///< Position in LHCb reference system
    Gaudi::SymMatrix3x3 m_covMatrix;  ///< Covariance matrix containing errors on vertex position
    double              m_chi2{-1.0}; ///< Chi square of vertex fit
    int                 m_nDoF{-1};   ///< Number of degree of freedom
    ExtraInfo           m_extraInfo;  ///< Some addtional user information. Don't use directly. Use *Info() methods.

  }; // class VertexBase

  /// Definition of Keyed Container for VertexBase
  typedef KeyedContainer<VertexBase, Containers::HashMap> VertexBases;

  inline std::ostream& operator<<( std::ostream& s, LHCb::VertexBase::VertexType e ) {
    switch ( e ) {
    case LHCb::VertexBase::Primary:
      return s << "Primary";
    default:
      return s << "ERROR wrong value " << int( e ) << " for enum LHCb::VertexBase::VertexType";
    }
  }

} // namespace LHCb

// -----------------------------------------------------------------------------
// end of class
// -----------------------------------------------------------------------------

// Including forward declarations

inline LHCb::VertexBase& LHCb::VertexBase::operator=( const VertexBase& rhs ) {

  if ( this != &rhs ) {
    m_position  = rhs.m_position;
    m_covMatrix = rhs.m_covMatrix;
    m_chi2      = rhs.m_chi2;
    m_nDoF      = rhs.m_nDoF;
    m_extraInfo = rhs.m_extraInfo;
  }
  return *this;
}

inline double LHCb::VertexBase::chi2PerDoF() const { return ( ( m_nDoF > 0 ) ? m_chi2 / double( m_nDoF ) : -1. ); }

inline void LHCb::VertexBase::setChi2AndDoF( double chi2, int ndof ) {
  m_chi2 = chi2;
  m_nDoF = ndof;
}

inline const LHCb::VertexBase::ExtraInfo& LHCb::VertexBase::extraInfo() const { return m_extraInfo; }

inline bool LHCb::VertexBase::hasInfo( int key ) const { return m_extraInfo.end() != m_extraInfo.find( key ); }

inline bool LHCb::VertexBase::addInfo( const int key, const double info ) {
  return m_extraInfo.insert( key, info ).second;
}

inline double LHCb::VertexBase::info( const int key, const double def ) const {
  auto i = m_extraInfo.find( key );
  return m_extraInfo.end() == i ? def : i->second;
}

inline unsigned long LHCb::VertexBase::eraseInfo( int key ) { return m_extraInfo.erase( key ); }

namespace Relations {
  template <typename>
  struct ObjectTypeTraits;

  template <>
  struct ObjectTypeTraits<LHCb::VertexBase> : KeyedObjectTypeTraits<LHCb::VertexBase> {};
} // namespace Relations
