###############################################################################
# (c) Copyright 2000-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
High level configuration tools for LHCb applications
"""
__version__ = "v17r0"
__author__ = "Marco Cattaneo <Marco.Cattaneo@cern.ch>"

__all__ = [
    'DstConf'  ## the configurable, configures DST writing/packing/unpacking
]

from Gaudi.Configuration import *
import GaudiKernel.ProcessJobOptions
from GaudiConf.CaloPackingConf import CaloDstPackConf, CaloDstUnPackConf
from Configurables import LHCbConfigurableUser as _LHCbConfigurableUser
from Configurables import RawEventFormatConf
from GaudiConf.SimConf import SimConf
from GaudiConf.DigiConf import DigiConf
from GaudiConf.TurboConf import TurboConf


class DummyWriter(_LHCbConfigurableUser):
    __slots__ = {"ItemList": [], "OptItemList": []}


class DstConf(_LHCbConfigurableUser):

    __slots__ = {
        "DstType": "NONE",
        "SimType": "None",
        "EnableUnpack": [],
        "EnablePackingChecks": False,
        "PackType": "TES",
        "PackSequencer": None,
        "Detectors": ['Velo', 'TT', 'IT', 'OT', 'Rich', 'Tr', 'Calo', 'Muon'],
        "Turbo": False,
        "Writer": "DstWriter",
        "OutputName": "",
        "SpilloverPaths": ["Prev", "PrevPrev", "Next", "NextNext"],
        "DataType": "",
        "WriteFSR": True,
        "SplitRawEventOutput": 2.0  #split raw event from Stripping 20
        ,
        "PVPersistence": "RecVertex"
    }

    _propertyDocDct = {
        'DstType':
        """ Type of dst, can be ['DST', 'RDST', 'XDST','MDST', 'LDST' ] """,
        'SimType':
        """ Type of simulation output, can be ['None','Minimal','Full'] """,
        'EnableUnpack':
        """ List of DST container types to be set up for on demand unpacking. See KnownUnpackingTypes """,
        'EnablePackingChecks':
        """ Flag to turn on the running of various unpacking checks, to test the quality of the data packing """,
        'PackType':
        """ Type of packing for the DST, can be ['NONE','TES','MDF'] """,
        'PackSequencer':
        """ Sequencer in which to run the packing algorithms """,
        'Turbo':
        """ Propagate Turbo stream locations """,
        'Writer':
        """ Name of OutputStream writing the DST """,
        'OutputName':
        """ Name of the output file, for MDF writing """,
        'SpilloverPaths':
        """ Paths to write to XDST if available on input file """,
        'DataType':
        """ Flag for backward compatibility with old data """,
        'WriteFSR':
        """ Flags whether to write out an FSR """,
        "SplitRawEventOutput":
        "Which version fo the split raw event to write out, default 2.0, same as Reco 14/Stripping 20 DSTs"
    }

    __used_configurables__ = [
        CaloDstPackConf,
        CaloDstUnPackConf,
        SimConf,
        DigiConf,
        TurboConf,
    ]

    KnownSimTypes = ['None', 'Minimal', 'Full']
    KnownDstTypes = ['NONE', 'DST', 'RDST', 'XDST', 'SDST', 'MDST', 'LDST']
    KnownPackTypes = ['NONE', 'TES', 'MDF']
    KnownUnpackingTypes = ["Reconstruction", "Stripping", "Tracking"]
    Run2DataTypes = ["2015", "2016", "2017", "2018"]

    def _doWrite(self, dType, pType, sType):
        """
        Define the file content and write it out
        """

        writer = DummyWriter()
        self._defineOutputData(dType, pType, sType, writer)

        if pType == 'MDF':
            if sType != 'None':
                raise TypeError("MDF packing not supported for Simulation")
            self._doWriteMDF(writer.ItemList)
        else:
            self._doWriteROOT(writer.ItemList, writer.OptItemList)

    def _defineOutputData(self, dType, pType, sType, writer):
        """
        Define content of the output dataset
        """

        # Choose whether to write packed or unpacked objects
        if pType == 'NONE':
            recDir = "Rec"
            if sType != "None":
                DigiConf().setProp("EnablePack", False)
                SimConf().setProp("EnablePack",
                                  DigiConf().getProp("EnablePack"))
        else:
            if not hasattr(self, "PackSequencer"):
                raise TypeError(
                    "Packing requested but PackSequencer not defined")
            recDir = "pRec"
            if sType != "None":
                DigiConf().setProp("EnablePack", True)
                self.setOtherProps(DigiConf(), ["PackSequencer"])
                SimConf().setProp("EnablePack",
                                  DigiConf().getProp("EnablePack"))

        # Add depth if not MDF
        if pType == 'MDF':
            depth = ""
        else:
            depth = "#1"

        writer.ItemList += [
            "/Event/Rec/Header" + depth, "/Event/Rec/Status" + depth
        ]
        from Configurables import LHCbApp

        writer.ItemList += ["/Event/Rec/Summary" + depth]

        writer.ItemList += ["/Event/" + recDir + "/Track/Best" + depth]

        if (self.getProp("PVPersistence") == "RecVertex"):
            writer.ItemList += ["/Event/" + recDir + "/Vertex/Primary" + depth]
            # for Run 2, fitted velo tracks are stored on DST to make PV refit possible offline
            if (self.getProp("DataType") in self.Run2DataTypes):
                writer.ItemList += [
                    "/Event/" + recDir + "/Track/FittedHLT1VeloTracks" + depth
                ]
        elif (self.getProp("PVPersistence") == "PrimaryVertex"):
            writer.ItemList += [
                "/Event/" + recDir + "/Vertex/PrimaryVertices" + depth
            ]
        else:
            raise RuntimeError("Unknown PVPersistence type: '%s'" %
                               self.getProp("PVPersistence"))

        if "Rich" in self.getProp("Detectors"):
            writer.ItemList += ["/Event/" + recDir + "/Rich/PIDs" + depth]
        if "Muon" in self.getProp("Detectors"):
            writer.ItemList += ["/Event/" + recDir + "/Muon/MuonPID" + depth]
        if "Calo" in self.getProp("Detectors"):
            writer.ItemList += [
                "/Event/" + recDir + "/Calo/Electrons" + depth,
                "/Event/" + recDir + "/Calo/Photons" + depth,
                "/Event/" + recDir + "/Calo/MergedPi0s" + depth,
                # "/Event/" + recDir + "/Calo/SplitPhotons" + depth
            ]

        writer.ItemList += [
            "/Event/" + recDir + "/ProtoP/Charged" + depth,
            "/Event/" + recDir + "/ProtoP/Neutrals" + depth,
            "/Event/" + recDir + "/Vertex/V0" + depth
        ]

        if "Muon" in self.getProp("Detectors"):
            writer.ItemList += ["/Event/" + recDir + "/Track/Muon" + depth]

        # Additional objects not packable as MDF
        if pType != "MDF":
            #Add the RawEvent, except for RDST. get RawEventLocations from the RawEventFormat database
            if dType != "RDST":
                #copied code from the RawEventJuggler
                RawEventFormatConf().loadIfRequired()
                from RawEventCompat.Configuration import _replaceWrap, _getDict, ReverseDict
                locs, rec = _getDict()
                output_locations = ReverseDict(
                    self.getProp("SplitRawEventOutput"), locs, rec)
                for loc in sorted(output_locations):
                    if not loc.startswith("/Event/"):
                        loc = ("/Event/" + loc).replace("//", "/")
                    writer.ItemList += [_replaceWrap(loc) + "#1"]

            # Add the simulation objects if simulation DST
            if sType != "None":

                eventLocations = ['']
                if dType == "XDST":
                    eventLocations = SimConf().allEventLocations()

                # Minimal MC output.
                SimConf().addHeaders(writer)
                SimConf().addMCVertices(writer, eventLocations)

                if self.getProp('Turbo'):
                    writer.OptItemList += TurboConf.optItemList()

                if sType == "Full":

                    writer.ItemList += [
                        # Links to from reconstructed objects to MCParticles
                        "/Event/Link/Rec/Track/Best#1"
                    ]
                    if "Calo" in self.getProp("Detectors"):
                        writer.ItemList += [
                            "/Event/Link/Rec/Calo/Photons#1",
                            "/Event/Link/Rec/Calo/Electrons#1",
                            "/Event/Link/Rec/Calo/MergedPi0s#1",
                            # "/Event/Link/Rec/Calo/SplitPhotons#1"
                        ]

                    # Objects propagated from Gauss
                    # Generation information
                    SimConf().addGenInfo(writer)
                    # MCparticles
                    SimConf().addMCParticles(writer, eventLocations)

                    # Objects propagated from Boole
                    # Digi headers
                    DigiConf().addHeaders(writer)
                    DigiConf().addMCTrackInfo(writer)
                    DigiConf().addMCCaloInfo(writer)

                    if dType == "XDST":
                        # Add the MCHits (from Gauss) and links to them (from Boole)
                        SimConf().addSubDetSimInfo(writer)
                        DigiConf().addMCHitLinks(writer)
                        DigiConf().addMCHitLinksSpillover(writer)
                        DigiConf().addMCParticleLinksSpillover(writer)
                    if dType in ["XDST", "LDST"]:
                        # Links from Digits to MCParticles
                        DigiConf().addMCParticleLinks(writer)
                        # Digitisation summaries
                        DigiConf().addMCDigitSummaries(writer)

    def _doWriteROOT(self, items, optItems):
        """
        Write a DST (or XDST) in ROOT format
        """
        writer = OutputStream(self.getProp("Writer"))
        writer.Preload = False
        writer.ItemList += items
        writer.OptItemList += optItems
        log.info("%s.ItemList=%s" % (self.getProp("Writer"), writer.ItemList))
        log.info(
            "%s.OptItemList=%s" % (self.getProp("Writer"), writer.OptItemList))
        from GaudiConf.IOHelper import IOHelper
        # Set a default file name if not already set
        if not writer.isPropertySet("Output"):
            outputFile = self.getProp("OutputName")
            outputFile = "PFN:" + outputFile + '.' + self.getProp(
                "DstType").lower()
        else:
            outputFile = IOHelper().undressFile(writer.getProp("Output"))
        # Add to the ApplicationMgr with corresct output persistency
        IOHelper().outStream(outputFile,
                             "OutputStream/" + self.getProp("Writer"),
                             self.getProp("WriteFSR"))

    def _doWriteMDF(self, items):
        """
        Write an DST in MDF format
        """
        from Configurables import WritePackedDst, LHCb__MDFWriter

        MDFpacker = WritePackedDst('MdfPacker')
        MDFpacker.Containers += items

        MDFwr = LHCb__MDFWriter('MdfWriter')
        MDFwr.Connection = self.getProp("OutputName") + '.mdf'
        MDFwr.Compress = 2
        MDFwr.GenerateMD5 = True
        MDFwr.BankLocation = '/Event/DAQ/DstEvent'

        GaudiSequencer("WriteMDFSeq").Members += [MDFpacker, MDFwr]

    def _doPack(self):
        """
        Set up the sequence to create the packed containers
        """
        packDST = self.getProp("PackSequencer")

        from Configurables import (TrackPacker, ProtoParticlePacker,
                                   RecVertexPacker, TwoProngVertexPacker,
                                   RichPIDPacker, MuonPIDPacker, VertexPacker,
                                   ChargedProtoParticleRemovePIDInfo)

        doChecks = self.getProp("EnablePackingChecks")

        # Pack tracks if they were not read from the input file...
        if "Tracking" not in self.getProp("EnableUnpack"):
            packDST.Members += [
                TrackPacker(name="PackTrack", EnableCheck=doChecks)
            ]
            # for Run2, add packed Velo tracks
            if ((self.getProp("DataType") in self.Run2DataTypes)
                    and (self.getProp("PVPersistence") == "RecVertex")):
                packFittedVelo = TrackPacker(
                    name="PackTracksFittedVelo",
                    InputName="Rec/Track/FittedHLT1VeloTracks",
                    OutputName="pRec/Track/FittedHLT1VeloTracks",
                    EnableCheck=doChecks)
                packDST.Members += [packFittedVelo]
            ###
        richpidpack = RichPIDPacker(name="PackRichPIDs", EnableCheck=doChecks)
        muonpidpack = MuonPIDPacker(name="PackMuonPIDs", EnableCheck=doChecks)

        if "Rich" in self.getProp("Detectors"):
            packDST.Members += [richpidpack]
        if "Muon" in self.getProp("Detectors"):
            packDST.Members += [muonpidpack]

        caloPackSeq = GaudiSequencer("CaloPacking")
        if "Calo" in self.getProp("Detectors"):
            packDST.Members += [caloPackSeq]

            caloPack = CaloDstPackConf()
            if not caloPack.isPropertySet('Enable'):
                CaloDstPackConf(Enable=True)
            caloPack.Sequence = caloPackSeq

        # Clean the PID information in the Charged ProtoParticles
        protoPidClean = ChargedProtoParticleRemovePIDInfo(
            "ProtoParticlePIDClean")
        packDST.Members += [protoPidClean]

        # packed the charged protos
        packChargedPs = ProtoParticlePacker(
            name="PackChargedProtos",
            InputName="/Event/Rec/ProtoP/Charged",
            OutputName="/Event/pRec/ProtoP/Charged",
            EnableCheck=doChecks)
        packDST.Members += [packChargedPs]

        # pack the neutral protos
        packNeutralPs = ProtoParticlePacker(
            name="PackNeutralProtos",
            InputName="/Event/Rec/ProtoP/Neutrals",
            OutputName="/Event/pRec/ProtoP/Neutrals",
            EnableCheck=doChecks)
        packDST.Members += [packNeutralPs]

        # Pack vertices if they were not read from the input file...
        if "Tracking" not in self.getProp("EnableUnpack"):
            if self.getProp("PVPersistence") == "RecVertex":
                packDST.Members += [RecVertexPacker()]
            elif self.getProp("PVPersistence") == "PrimaryVertex":
                packDST.Members += [VertexPacker()]
            packDST.Members += [TwoProngVertexPacker()]

        if "Muon" in self.getProp("Detectors"):
            packDST.Members += [
                TrackPacker(
                    name="PackMuonTracks",
                    InputName="/Event/Rec/Track/Muon",
                    OutputName="/Event/pRec/Track/Muon",
                    EnableCheck=doChecks)
            ]

        # In MDF case, add a sub sequence for the MDF writing
        if self.getProp("PackType").upper() == "MDF":
            packDST.Members += [GaudiSequencer("WriteMDFSeq")]

        # Tests
        if self.getProp("EnablePackingChecks"): self._doPackChecks()

    def _doPackChecks(self):

        packDST = self.getProp("PackSequencer")
        from Configurables import GaudiSequencer
        checks = GaudiSequencer("PackingChecks")
        packDST.Members += [checks]

        from Configurables import (UnpackRecVertex, UnpackTwoProngVertex)
        from Configurables import (CompareRecVertex, CompareTwoProngVertex)

        # Unpack to temporary locations
        tempLoc = "Test"
        unpackVertex = UnpackRecVertex("UnpackVertexTest")
        unpackVertexTest = UnpackRecVertex(
            "UnpackVertexTest",
            OutputName=str(unpackVertex.OutputName) + tempLoc)
        unpackV0 = UnpackTwoProngVertex("UnpackV0")
        unpackV0Test = UnpackTwoProngVertex(
            "UnpackV0Test", OutputName=str(unpackV0.OutputName) + tempLoc)
        checks.Members += [
            unpackVertex, unpackV0, unpackVertexTest, unpackV0Test
        ]

        # Comparisons
        checkVertex = CompareRecVertex(
            "CheckPackedVertices",
            InputName=unpackVertex.OutputName,
            TestName=unpackVertexTest.OutputName)
        checkV0 = CompareTwoProngVertex(
            "CheckPackedV0s",
            InputName=unpackV0.OutputName,
            TestName=unpackV0Test.OutputName)

        checks.Members += [checkVertex, checkV0]

    def _doUnpack(self):
        """
        Set up DataOnDemandSvc to unpack reconstruction information
        """
        from Configurables import (UnpackTrack, UnpackProtoParticle,
                                   UnpackRecVertex, UnpackTwoProngVertex,
                                   UnpackPrimaryVertices, UnpackWeightsVector,
                                   RecToPVConverter, PVToRecConverter)

        log.debug("In DstConf._doUnpack")

        DataOnDemandSvc().AlgMap["/Event/Rec/Track/Best"] = UnpackTrack()
        DataOnDemandSvc(
        ).AlgMap["/Event/Rec/Vertex/V0"] = UnpackTwoProngVertex()

        if (self.getProp("PVPersistence") == "RecVertex"):
            DataOnDemandSvc(
            ).AlgMap["/Event/Rec/Vertex/Primary"] = UnpackRecVertex()
            pvWunpack = UnpackWeightsVector("UnpackPVWeights")
            DataOnDemandSvc().AlgMap["/Event/Rec/Vertex/Weights"] = pvWunpack
            DataOnDemandSvc(
            ).AlgMap["/Event/Rec/Vertex/PrimaryVertices"] = RecToPVConverter()
        elif (self.getProp("PVPersistence") == "PrimaryVertex"):
            DataOnDemandSvc().AlgMap[
                "/Event/Rec/Vertex/PrimaryVertices"] = UnpackPrimaryVertices()
            DataOnDemandSvc(
            ).AlgMap["/Event/Rec/Vertex/Primary"] = PVToRecConverter()
        else:
            raise RuntimeError("Unknown PVPersistence type: '%s'" %
                               self.getProp("PVPersistence"))

        if (self.getProp("DataType") in self.Run2DataTypes):
            unpackFittedVeloTracks = UnpackTrack("unpackFittedVeloTracks")
            unpackFittedVeloTracks.InputName = "pRec/Track/FittedHLT1VeloTracks"
            unpackFittedVeloTracks.OutputName = "Rec/Track/FittedHLT1VeloTracks"
            DataOnDemandSvc().AlgMap[
                "/Event/Rec/Track/FittedHLT1VeloTracks"] = unpackFittedVeloTracks

        if "Tracking" in self.getProp("EnableUnpack"): return  # skip the rest

        # Calo
        if "Calo" in self.getProp("Detectors"):
            caloUnpack = CaloDstUnPackConf()
            if not caloUnpack.isPropertySet('Enable'):
                log.debug("Setting caloUnpack.Enable = True")
                caloUnpack.Enable = True
            else:
                log.debug("Not setting caloUnpack.Enable. Current value = %s",
                          caloUnpack.Enable)

        # RichPIDs
        if "Rich" in self.getProp("Detectors"):
            self._unpackRichPIDs()

        # MuonPIDs
        if "Muon" in self.getProp("Detectors"):
            self._unpackMuonPIDs()

        # ProtoParticles
        self._unpackProtoParticles()

        # Muon Tracks
        if "Muon" in self.getProp("Detectors"):
            unpackMuons = UnpackTrack(
                name="UnpackMuonTracks",
                OutputName="/Event/Rec/Track/Muon",
                InputName="/Event/pRec/Track/Muon")
            DataOnDemandSvc().AlgMap["/Event/Rec/Track/Muon"] = unpackMuons

    def _unpackMuonPIDs(self):

        from Configurables import (GaudiSequencer, UnpackMuonPIDs)

        mpidLoc = "/Event/Rec/Muon/MuonPID"

        mpidSeq = GaudiSequencer("UnpackMuonPIDSeq")
        DataOnDemandSvc().AlgMap[mpidLoc] = mpidSeq

        # Unpacking alg
        unpackmuonpid = UnpackMuonPIDs("UnpackMuonPIDs")
        mpidSeq.Members += [unpackmuonpid]

        # Additional processing, not for MDST
        inputtype = self.getProp('DstType').upper()
        if inputtype != 'MDST':

            # For backwards compat, also run an alg that if need be recreates MuonPIDs
            # from ProtoParticles. Can eventually be removed.
            from Configurables import MuonPIDsFromProtoParticlesAlg
            mpidSeq.Members += [MuonPIDsFromProtoParticlesAlg("CheckMuonPIDs")]

    def _unpackRichPIDs(self):

        from Configurables import (GaudiSequencer, UnpackRichPIDs)

        rpidLoc = "/Event/Rec/Rich/PIDs"

        rpidSeq = GaudiSequencer("UnpackRichPIDSeq")
        DataOnDemandSvc().AlgMap[rpidLoc] = rpidSeq

        # Alg that unpacks data from packed data
        unpackrichpid = UnpackRichPIDs("UnpackRichPIDs")
        rpidSeq.Members += [unpackrichpid]

        # Additional processing, not for MDST
        inputtype = self.getProp('DstType').upper()
        if inputtype != 'MDST':

            # For backwards compat, also run an alg that if need be recreates RichPIDs
            # from ProtoParticles. Can eventually be removed.
            from Configurables import RichPIDsFromProtoParticlesAlg
            rpidSeq.Members += [RichPIDsFromProtoParticlesAlg("CheckRichPIDs")]

    def _unpackProtoParticles(self):

        from Configurables import (GaudiSequencer, UnpackProtoParticle)

        # Neutrals
        # --------

        neutralLoc = "/Event/Rec/ProtoP/Neutrals"
        unpackNeutrals = UnpackProtoParticle(
            name="UnpackNeutralProtos",
            OutputName=neutralLoc,
            InputName="/Event/pRec/ProtoP/Neutrals")
        DataOnDemandSvc().AlgMap[neutralLoc] = unpackNeutrals

        # Charged
        # -------

        chargedLoc = "/Event/Rec/ProtoP/Charged"
        chargedSeq = GaudiSequencer("UnpackChargedProtosSeq")
        DataOnDemandSvc().AlgMap[chargedLoc] = chargedSeq

        # Unpacker
        unpackCharged = UnpackProtoParticle(
            name="UnpackChargedProtos",
            OutputName=chargedLoc,
            InputName="/Event/pRec/ProtoP/Charged")
        chargedSeq.Members += [unpackCharged]

        # Additional processing, not for MDST
        inputtype = self.getProp('DstType').upper()
        if inputtype != 'MDST':

            # PID calibration
            from Configurables import (
                ChargedProtoParticleAddInfo, ChargedProtoParticleAddRichInfo,
                ChargedProtoParticleAddMuonInfo,
                ChargedProtoParticleAddCombineDLLs, TESCheck)
            recalib = GaudiSequencer("ProtoParticleCombDLLs")
            recalib.IgnoreFilterPassed = False
            chargedSeq.Members += [recalib]
            # Filter to check input Protos exist
            recalib.Members += [
                TESCheck(
                    name="CheckChargedProtosExist",
                    Inputs=[chargedLoc],
                    OutputLevel=5,
                    Stop=False)
            ]
            # Add Rich and Muon PID results, Combined DLLs to protoparticles
            recalib.Members += [
                ChargedProtoParticleAddInfo(
                    "ChargedProtoAddInfo",
                    AddInfo=[
                        ChargedProtoParticleAddMuonInfo(
                            "ChargedProtoPAddMuon"),
                        ChargedProtoParticleAddRichInfo(
                            "ChargedProtoPAddRich"),
                        ChargedProtoParticleAddCombineDLLs(
                            "ChargedProtoPAddCombDLLs")
                    ])
            ]

    def _unpackStripping(self):

        from Configurables import (ConversionDODMapper,
                                   ParticlesAndVerticesMapper,
                                   TrackClustersMapper)

        # General unpacking mapping
        mapper = ConversionDODMapper("UnpackRecPhysMapper")
        # The input <-> output mappings
        mapper.Transformations = [('(.*)/Rec(.*)', '$1/pRec$2'),
                                  ('(.*)/Raw(.*)', '$1/pRaw$2'),
                                  ('(.*)/Phys(.*)', '$1/pPhys$2'),
                                  ('(.*)/MC(.*)', '$1/pMC$2')]
        # algorithm types from source ClassIDs
        mapper.Algorithms[1550] = "UnpackTrack"
        mapper.Algorithms[1552] = "UnpackProtoParticle"
        if "Calo" in self.getProp("Detectors"):
            mapper.Algorithms[1551] = "UnpackCaloHypo"
            mapper.Algorithms[
                1541] = "DataPacking::Unpack<LHCb::CaloClusterPacker>"
            mapper.Algorithms[
                1542] = "DataPacking::Unpack<LHCb::CaloDigitPacker>"
            mapper.Algorithms[
                1543] = "DataPacking::Unpack<LHCb::CaloAdcPacker>"
        if "Rich" in self.getProp("Detectors"):
            mapper.Algorithms[
                1561] = "DataPacking::Unpack<LHCb::RichPIDPacker>"
        if "Muon" in self.getProp("Detectors"):
            mapper.Algorithms[
                1571] = "DataPacking::Unpack<LHCb::MuonPIDPacker>"
        mapper.Algorithms[1553] = "UnpackRecVertex"
        mapper.Algorithms[
            1555] = "DataPacking::Unpack<LHCb::WeightsVectorPacker>"
        mapper.Algorithms[1581] = "UnpackParticlesAndVertices"
        mapper.Algorithms[1559] = "UnpackDecReport"
        mapper.Algorithms[1510] = "UnpackMCParticle"
        mapper.Algorithms[1511] = "UnpackMCVertex"

        # Packed Particles and Vertices
        pvmapper = ParticlesAndVerticesMapper("UnpackPsAndVsMapper")
        # Turbo unpacking requires a job configured for Turbo data processing,
        # and fails otherwise. Here we prevent the unpacking from being
        # triggered if this isn't a Turbo job
        if not self.getProp("Turbo"):
            pvmapper.VetoStreams.append("/Event/Turbo")

        # Cluster upacking
        clusmapper = TrackClustersMapper("UnpackTkClustersMapper")
        # Lite cluster creation
        lclusmapper = ConversionDODMapper("LiteClusterMapper")
        lclusmapper.InputOptionName = "inputLocation"
        lclusmapper.OutputOptionName = "outputLocation"
        lclusmapper.Transformations = [('(.*)/Raw/(.*)/LiteClusters',
                                        '$1/Raw/$2/Clusters')]
        if "Velo" in self.getProp("Detectors"):
            lclusmapper.Algorithms[397222] = "VeloClustersToLite"
        lclusmapper.Algorithms[402220] = "STClustersToLite"

        # Add the tools to the DOD service tools lists
        tools = [clusmapper, lclusmapper, pvmapper, mapper]
        DataOnDemandSvc().NodeMappingTools += tools
        DataOnDemandSvc().AlgMappingTools += tools

    def __apply_configuration__(self):

        log.info(self)

        sType = self.getProp("SimType").capitalize()
        dType = self.getProp("DstType").upper()

        if sType not in self.KnownSimTypes:
            raise TypeError("Unknown SimType '%s'" % sType)
        if sType != "None":
            if dType == 'RDST':
                raise TypeError("Cannot have RDST for simulation!")
            DigiConf().setProp("EnableUnpack", True)
            SimConf().setProp("EnableUnpack",
                              DigiConf().getProp("EnableUnpack"))

        if dType not in self.KnownDstTypes:
            raise TypeError("Unknown DstType '%s'" % dType)
        if dType == "SDST":
            log.warning("SDST output type is obsolete, setting it to DST")
            dtype = "DST"

        # Propagate SpilloverPaths and DataType to DigiConf and to SimConf via DigiConf
        self.setOtherProps(DigiConf(),
                           ["SpilloverPaths", "DataType", "Detectors"])
        DigiConf().setOtherProps(SimConf(), ["SpilloverPaths", "DataType"])

        pType = self.getProp("PackType").upper()
        if pType not in self.KnownPackTypes:
            raise TypeError("Unknown PackType '%s'" % pType)

        # Unpacking options
        for unpackT in self.getProp("EnableUnpack"):
            if unpackT not in self.KnownUnpackingTypes:
                raise TypeError("Unknown Unpacking type '%s'" % unpackT)
        if "Reconstruction" in self.getProp(
                "EnableUnpack") or "Tracking" in self.getProp("EnableUnpack"):
            self._doUnpack()
        if "Stripping" in self.getProp("EnableUnpack"): self._unpackStripping()

        if dType != 'NONE':
            if hasattr(self, "PackSequencer"): self._doPack()
            GaudiKernel.ProcessJobOptions.PrintOn(force=True)
            self._doWrite(dType, pType, sType)
            GaudiKernel.ProcessJobOptions.PrintOff()
