##############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Helpers for configuring an application to read Moore HLT2 output."""
from __future__ import absolute_import
import os
from PyConf.application import (
    default_raw_event, make_data_with_FetchDataFromFile, ComponentConfig)
from PyConf.dataflow import dataflow_config
from PyConf.components import force_location, setup_component
from PyConf.reading import get_mc_vertices, get_mc_particles

__all__ = ["do_unpacking", "type_map"]


# These are the known types that we can persist
def type_map():
    ## TODO: derive from the properties of SelectivePacker (which doesn't work trivially, as there is no DataHandleArray (and corresponding property) in Gaudi...)
    type_map = {
        "SharedObjectsContainer<LHCb::Particle>":
        "ParticlesSelection",
        "SharedObjectsContainer<LHCb::ProtoParticle>":
        "ProtoParticlesSelection",
        "SharedObjectsContainer<LHCb::Event::v1::Track>":
        "TracksSelection",
        "SharedObjectsContainer<LHCb::RichPID>":
        "RichPIDsSelection",
        "SharedObjectsContainer<LHCb::MuonPID>":
        "MuonPIDsSelection",
        "SharedObjectsContainer<LHCb::RecVertex>":
        "PVsSelection",
        "SharedObjectsContainer<LHCb::Vertex>":
        "VerticesSelection",
        "SharedObjectsContainer<LHCb::TwoProngVertex>":
        "TwoProngVerticesSelection",
        "SharedObjectsContainer<LHCb::CaloHypo>":
        "CaloHyposSelection",
        "SharedObjectsContainer<LHCb::CaloCluster>":
        "CaloClustersSelection",
        "SharedObjectsContainer<LHCb::FlavourTag>":
        "FlavourTagsSelection",
        "KeyedContainer<LHCb::Particle,Containers::KeyedObjectManager<Containers::hashmap> >":
        "Particles",
        "KeyedContainer<LHCb::ProtoParticle,Containers::KeyedObjectManager<Containers::hashmap> >":
        "ProtoParticles",
        "KeyedContainer<LHCb::Event::v1::Track,Containers::KeyedObjectManager<Containers::hashmap> >":
        "Tracks",
        "KeyedContainer<LHCb::RichPID,Containers::KeyedObjectManager<Containers::hashmap> >":
        "RichPIDs",
        "KeyedContainer<LHCb::CaloChargedPID,Containers::KeyedObjectManager<Containers::hashmap> >":
        "CaloChargedPIDs",
        "KeyedContainer<LHCb::BremInfo,Containers::KeyedObjectManager<Containers::hashmap> >":
        "BremInfos",
        "KeyedContainer<LHCb::MuonPID,Containers::KeyedObjectManager<Containers::hashmap> >":
        "MuonPIDs",
        "KeyedContainer<LHCb::GlobalChargedPID,Containers::KeyedObjectManager<Containers::hashmap> >":
        "GlobalChargedPIDs",
        "KeyedContainer<LHCb::NeutralPID,Containers::KeyedObjectManager<Containers::hashmap> >":
        "NeutralPIDs",
        "KeyedContainer<LHCb::RecVertex,Containers::KeyedObjectManager<Containers::hashmap> >":
        "PVs",
        "KeyedContainer<LHCb::Vertex,Containers::KeyedObjectManager<Containers::hashmap> >":
        "Vertices",
        "KeyedContainer<LHCb::TwoProngVertex,Containers::KeyedObjectManager<Containers::hashmap> >":
        "TwoProngVertices",
        "KeyedContainer<LHCb::CaloHypo,Containers::KeyedObjectManager<Containers::hashmap> >":
        "CaloHypos",
        "KeyedContainer<LHCb::CaloCluster,Containers::KeyedObjectManager<Containers::hashmap> >":
        "CaloClusters",
        "KeyedContainer<LHCb::CaloDigit,Containers::KeyedObjectManager<Containers::hashmap> >":
        "CaloDigits",
        "KeyedContainer<LHCb::CaloAdc,Containers::KeyedObjectManager<Containers::hashmap> >":
        "CaloAdcs",
        "KeyedContainer<LHCb::FlavourTag,Containers::KeyedObjectManager<Containers::hashmap> >":
        "FlavourTags",
        "KeyedContainer<LHCb::WeightsVector,Containers::KeyedObjectManager<Containers::hashmap> >":
        "WeightsVectors",
        "LHCb::Relation1D<LHCb::Particle,LHCb::VertexBase>":
        "P2VRelations",
        "LHCb::Relation1D<LHCb::Particle,LHCb::MCParticle>":
        "P2MCPRelations",
        "LHCb::Relation1D<LHCb::Particle,int>":
        "P2IntRelations",
        "LHCb::Relation1D<LHCb::Particle,LHCb::RelatedInfoMap>":
        "P2InfoRelations",
        "LHCb::RelationWeighted1D<LHCb::ProtoParticle,LHCb::MCParticle,double>":
        "PP2MCPRelations",
        "LHCb::RecSummary":
        "RecSummary",
        "LHCb::Event::PV::PrimaryVertexContainer":
        "ExtendedPVs",
    }

    return type_map


def __bank_location(input_process="Hlt2",
                    stream="default",
                    bank="DstData",
                    raw_event_format=0.5):

    if input_process in ["Spruce", "Turbo"]:
        bank_location = make_data_with_FetchDataFromFile(
            f'/Event/{stream}/RawEvent', force_type='LHCb::RawEvent')
    else:
        bank_location = default_raw_event([bank],
                                          raw_event_format=raw_event_format)
    return bank_location


def __unpack_rawevent(bank_types=["DstData", "HltDecReports"],
                      input_process='Hlt2',
                      stream="default",
                      raw_event_format=0.5):
    """Return RawBank:Views of banks in RawEvent.

    Args:
        bank_types (list of str): RawBanks to create RawBank:Views of.
        input_process (str): 'Turbo' or 'Spruce' or 'Hlt2' - serves to determine `RawEventLocation` only.
        stream (str): needed post-sprucing as RawEvent is then dependent on stream name
            - drives `RawEventLocation` only.
        raw_event_format (float):
    """
    assert input_process in [
        "Spruce", "Turbo", "Hlt2"
    ], 'Unpacking helper only accepts Turbo, Spruce or Hlt2 processes'

    bank_types = sorted(set(bank_types))

    from PyConf.Algorithms import LHCb__UnpackRawEvent
    unpackrawevent = LHCb__UnpackRawEvent(
        BankTypes=bank_types,
        RawBankLocations=[
            force_location(f"DAQ/RawBanks/{rb}") for rb in bank_types
        ],
        RawEventLocation=__bank_location(input_process, stream, "DstData",
                                         raw_event_format))
    return __as_configurables(unpackrawevent)


def __decoder(input_process='Hlt2'):
    """Return a DstData raw bank decoder.

    Args:
        input_process (str): 'Turbo' or 'Spruce' or 'Hlt2' - serves to determine `RawEventLocation` only.
    """
    assert input_process in [
        "Spruce", "Turbo", "Hlt2"
    ], 'Unpacking helper only accepts Turbo, Spruce or Hlt2 processes'
    from PyConf.Algorithms import HltPackedBufferDecoder
    decoder = HltPackedBufferDecoder(
        SourceID=input_process.replace("Turbo", "Hlt2"),
        DecReports=force_location("DAQ/RawBanks/HltDecReports"),
        RawBanks=force_location("DAQ/RawBanks/DstData"))
    return __as_configurables(decoder)


def __hlt_decisions(source='Hlt2'):
    """Return a HltDecReportsDecoder instance for HLT1, HLT2, Spruce or Turbo decisions.

    Args:
        source (str): 'Hlt1' or 'Hlt2' or 'Spruce'- indicates whether selection stage is HLT1, HLT2 or Spruce
    """

    from PyConf.Algorithms import HltDecReportsDecoder
    decode_hlt = HltDecReportsDecoder(
        name=source + "DecReportsDecoder",
        SourceID=source,
        RawBanks=force_location("DAQ/RawBanks/HltDecReports"),
        outputs={
            'OutputHltDecReportsLocation':
            force_location(f'/Event/{source}/DecReports')
        },
    )
    return __as_configurables(decode_hlt)


def do_unpacking(stream="default",
                 input_process='Hlt2',
                 simulation=True,
                 raw_event_format=0.5):
    """Return a list of algorithm that will do a full unpacking of data

    Args:
        input_process (str): 'Turbo' or 'Spruce' or 'Hlt2' - serves to determine `RawEventLocation` only.
        stream (str): needed post-sprucing as RawEvent is then dependent on stream name
            - drives `RawEventLocation` only.
        raw_event_format (float):
    """
    TES_ROOT = '/Event/HLT2'
    if input_process == "Spruce":
        TES_ROOT = '/Event/Spruce'

    algs = []
    unpack = __unpack_rawevent(
        bank_types=["DstData", "HltDecReports", "HltRoutingBits"],
        input_process=input_process,
        stream=stream,
        raw_event_format=raw_event_format)
    algs += [unpack]

    # Get the hlt2 decisions
    algs += [__hlt_decisions(source="Hlt2")]

    # if sprucing, get sprucing decisions
    if input_process == "Spruce" or input_process == "Turbo":
        algs += [__hlt_decisions(source="Spruce")]

    # decode raw banks to map of packed data buffers
    packed_decoder = __decoder(input_process=input_process)
    algs += [packed_decoder]

    # get mc unpackers (shouldn't this be only for simulation?)
    if simulation:
        assert input_process in [
            "Hlt2", "Turbo", "Spruce"
        ], 'MC unpacking only accepts Turbo, Spruce or Hlt2 process'
        prefix = 'Spruce' if input_process == "Spruce" else ''
        algs += __as_configurables([
            get_mc_vertices(
                os.path.join("/Event", prefix, "HLT2/MC/Vertices")).producer,
            get_mc_particles(
                os.path.join("/Event", prefix, "HLT2/MC/Particles")).producer
        ])

    # Finally unpack packed data buffers to objects
    from Configurables import UnpackDstDataBank
    algs += [UnpackDstDataBank(InputName=str(packed_decoder.OutputBuffers))]

    if input_process == "Hlt2":
        from Configurables import HltRoutingBitsFilter, GaudiSequencer
        phys_filter = HltRoutingBitsFilter(
            name="PhysFilter",
            RawBanks="DAQ/RawBanks/HltRoutingBits",
            RequireMask=(0, 0, 1 << (95 - 64)))
        algs = [
            unpack,
            GaudiSequencer(
                "UnpackAfterPhysFilter", Members=[phys_filter] + algs[1:])
        ]
    return algs


def __as_configurables(algs):
    """Convert a list of PyConf Algorithm objects to Configurables.

    The returned list is not guaranteed to be in an order which satisfies
    the data dependencies of the algorithms when run sequentially. Any
    tools the Algorithm objects reference will be ignored.

    This is required for the GaudiPython tests (PyConf uses HLT scheduler which has no support for GP),
    ie. we need Configurables not PyConf Algorithm objects
    """
    configuration = dataflow_config()
    if isinstance(algs, list):
        for alg in algs:
            configuration.update(alg.configuration())
        configurables, _ = configuration.apply()
        return configurables
    else:
        configuration.update(algs.configuration())
        configurables, _ = configuration.apply()

        for c in configurables:
            if c.name() == algs.name: return c
