/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

//-----------------------------------------------------------------------------
/** @file RichRadIntersection.h
 *
 *  Header file for RICH utility class : Rich::RadIntersection
 *
 *  @author Antonis Papanestis
 *  @date   2006-02-28
 */
//-----------------------------------------------------------------------------

#pragma once

// STL
#include <algorithm>
#include <any>
#include <cmath>
#include <functional>
#include <utility>

// Include files
#include "GaudiKernel/Point3DTypes.h"
#include "GaudiKernel/Vector3DTypes.h"

// Boost
#include <boost/container/small_vector.hpp>

namespace Rich {

  /** @class RadIntersection RichRadIntersection.h Kernel/RichRadIntersection.h
   *
   *  RICH Radiator intersection information.
   *
   *  Contains the entry and exit points of a single intersection with a radiator
   *  volume, together with the momentum vector at the intersection points, and
   *  a pointer to the intersected radiator volume.
   *
   *  @author Antonis Papanestis
   *  @author Chris Jones   Christopher.Rob.Jones@cern.ch
   *  @date   2006-02-28
   */

  class RadIntersection {

  public:
    /// Definition of a vector of intersections
    using Vector = boost::container::small_vector<Rich::RadIntersection, 1>;

  public:
    /// Default constructor
    RadIntersection() = default;

    /// Default destructor
    ~RadIntersection() = default;

    /// Default Copy Constructor
    RadIntersection( const RadIntersection& ) = default;

    /// Default Copy Operator
    RadIntersection& operator=( const RadIntersection& ) = default;

    /// Default Move Constructor
    RadIntersection( RadIntersection&& ) = default;

    /// Default Move Operator
    RadIntersection& operator=( RadIntersection&& ) = default;

  public:
    /// Constructor from (copied) entry and exit points and momentum vectors, and a pointer to a
    /// radiator object
    template <typename RADIATOR>
    RadIntersection( Gaudi::XYZPoint  entryPoint, ///< The radiator entry point
                     Gaudi::XYZVector entryVect,  ///< The direction vector at the entry point
                     Gaudi::XYZPoint  exitPoint,  ///< The radiator exit point
                     Gaudi::XYZVector exitVect,   ///< The direction vector at the exit point
                     const RADIATOR*  rad         ///< Pointer to the intersected radiator volume
                     )
        : m_entryPoint( std::move( entryPoint ) )
        , m_entryVect( std::move( entryVect ) )
        , m_exitPoint( std::move( exitPoint ) )
        , m_exitVect( std::move( exitVect ) )
        , m_radiator( rad ) {}

    /// Constructor from (moved) entry and exit points and momentum vectors, and a pointer to a
    /// radiator object
    template <typename RADIATOR>
    RadIntersection( Gaudi::XYZPoint&&  entryPoint, ///< The radiator entry point
                     Gaudi::XYZVector&& entryVect,  ///< The direction vector at the entry point
                     Gaudi::XYZPoint&&  exitPoint,  ///< The radiator exit point
                     Gaudi::XYZVector&& exitVect,   ///< The direction vector at the exit point
                     const RADIATOR*    rad         ///< Pointer to the intersected radiator volume
                     )
        : m_entryPoint( std::move( entryPoint ) )
        , m_entryVect( std::move( entryVect ) )
        , m_exitPoint( std::move( exitPoint ) )
        , m_exitVect( std::move( exitVect ) )
        , m_radiator( rad ) {}

  public:
    /// Set the entry point to the radiator
    inline void setEntryPoint( const Gaudi::XYZPoint& entryPoint ) noexcept { m_entryPoint = entryPoint; }

    /// Get the entry point to the radiator (read only)
    [[nodiscard]] inline const Gaudi::XYZPoint& entryPoint() const noexcept { return m_entryPoint; }

    /// Get the entry point to the radiator (read/write)
    [[nodiscard]] inline Gaudi::XYZPoint& entryPoint() noexcept { return m_entryPoint; }

    /// Set the entry momentum vector to the radiator
    inline void setEntryMomentum( const Gaudi::XYZVector& entryV ) noexcept { m_entryVect = entryV; }

    /// Get the entry point to the radiator (read only)
    [[nodiscard]] inline const Gaudi::XYZVector& entryMomentum() const noexcept { return m_entryVect; }

    /// Get the entry point to the radiator (read/write)
    [[nodiscard]] inline Gaudi::XYZVector& entryMomentum() noexcept { return m_entryVect; }

    /// Set the exit point of the radiator
    inline void setExitPoint( const Gaudi::XYZPoint& exitPoint ) noexcept { m_exitPoint = exitPoint; }

    /// Get the exit point to the radiator (read only)
    [[nodiscard]] inline const Gaudi::XYZPoint& exitPoint() const noexcept { return m_exitPoint; }

    /// Get the exit point to the radiator (read/write)
    [[nodiscard]] inline Gaudi::XYZPoint& exitPoint() noexcept { return m_exitPoint; }

    /// Set the exit momentum vector to the radiator
    inline void setExitMomentum( const Gaudi::XYZVector& exitV ) noexcept { m_exitVect = exitV; }

    /// Get the entry point to the radiator (read only)
    [[nodiscard]] inline const Gaudi::XYZVector& exitMomentum() const noexcept { return m_exitVect; }

    /// Get the entry point to the radiator (read/write)
    [[nodiscard]] inline Gaudi::XYZVector& exitMomentum() noexcept { return m_exitVect; }

    /// Set pointer to associated radiator detector element
    template <typename RADIATOR>
    inline void setRadiator( const RADIATOR* radiator ) noexcept {
      m_radiator = radiator;
    }

    /// Get pointer to associated radiator detector element as given type
    template <typename RADIATOR>
    [[nodiscard]] inline decltype( auto ) radiator() const noexcept {
      return std::any_cast<const RADIATOR*>( m_radiator );
    }

    /// Returns the path length (squared) in the given radiator
    [[nodiscard]] inline double pathLength2() const { return ( exitPoint() - entryPoint() ).Mag2(); }

    /// Returns the path length in the given radiator
    [[nodiscard]] inline double pathLength() const { return std::sqrt( pathLength2() ); }

    /// Overload output to ostream
    friend inline std::ostream& operator<<( std::ostream& os, const Rich::RadIntersection& intersect ) {
      os << "["
         << " Entry: Point=" << intersect.entryPoint() << " Dir=" << intersect.entryMomentum()
         << " Exit: Point=" << intersect.exitPoint() << " Dir=" << intersect.exitMomentum() << " ]";
      return os;
    }

  private:
    /// Entry point for radiator
    Gaudi::XYZPoint m_entryPoint;

    /// Momentum (direction) vector at the entry point
    Gaudi::XYZVector m_entryVect;

    /// Exit point for radiator
    Gaudi::XYZPoint m_exitPoint;

    /// Momentum (direction) vector at the entry point
    Gaudi::XYZVector m_exitVect;

    /** pointer to radiator detector
     *  NOTE: As long as we need to support DD4Hep and the old RichDet
     *        use void* here to allow storing either type.
     *        to be replaced with Detectors::Radiator once RichDet is fully dropped */
    std::any m_radiator;

  public:
    /** @class Sorter RichRadIntersection.h Kernel/RichRadIntersection.h
     *
     *  Utility class to sort the Radiator Intersections
     *
     *  @author Chris Jones   Christopher.Rob.Jones@cern.ch
     *  @date   14/10/2006
     */
    class Sorter {
    private: // utility classes
      /// Functor to sort RichSmartIDs by Rich then panel numbers
      struct SortByZ {

        /** Sort operator for the RadIntersection
         *
         *  Sorts into order of increasing z position
         *
         *  @param i1 First RadIntersection
         *  @param i2 Second RadIntersection
         *
         *  @return bool indicating if i1 should be listed before i2
         */
        inline bool operator()( const Rich::RadIntersection& i1, const Rich::RadIntersection& i2 ) const {
          return ( i1.entryPoint().z() < i2.entryPoint().z() );
        }
      };

    public:
      /// Constructor
      Sorter() = default;

      /// Destructor
      ~Sorter() = default;

      /// Sort the vector of intersections into order increasing Z position
      inline static void sortByZ( Rich::RadIntersection::Vector& intersections ) {
        std::stable_sort( intersections.begin(), intersections.end(), Rich::RadIntersection::Sorter::SortByZ() );
      }
    };
  };

} // namespace Rich
