/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Kernel/LHCbID.h"
#include "Kernel/RichSmartID.h"
#include <cstdint>
#include <iostream>

using namespace std;

int main( int /*argc*/, char** /*argv*/ ) {

  bool OK = true;

  // test HPDs
  for ( unsigned int irich = 0; irich <= LHCb::RichSmartID::HPD::MaxRich; ++irich ) {
    const auto rich = (Rich::DetectorType)irich;
    for ( unsigned int ipanel = 0; ipanel <= LHCb::RichSmartID::HPD::MaxPanel; ++ipanel ) {
      const auto panel = (Rich::Side)ipanel;
      for ( unsigned int pdrow = 0; pdrow <= LHCb::RichSmartID::HPD::MaxPDNumInCol; ++pdrow ) {
        for ( unsigned int pdcol = 0; pdcol <= LHCb::RichSmartID::HPD::MaxPDCol; ++pdcol ) {
          for ( unsigned int pixrow = 0; pixrow <= LHCb::RichSmartID::HPD::MaxPixelRow; ++pixrow ) {
            for ( unsigned int pixcol = 0; pixcol <= LHCb::RichSmartID::HPD::MaxPixelCol; ++pixcol ) {
              for ( unsigned int subpix = 0; subpix <= LHCb::RichSmartID::HPD::MaxPixelSubRow; ++subpix ) {

                // Make a test smart ID
                const LHCb::RichSmartID id( rich, panel, pdrow, pdcol, pixrow, pixcol, subpix,
                                            LHCb::RichSmartID::HPDID );

                // Check values
                if ( rich != id.rich() ||        //
                     panel != id.panel() ||      //
                     pdrow != id.pdNumInCol() || //
                     pdcol != id.pdCol() ||      //
                     pixrow != id.pixelRow() ||  //
                     pixcol != id.pixelCol() ||  //
                     subpix != id.pixelSubRow() ) {
                  cerr << "Problem with RichSmartID " << id << endl;
                  OK = false;
                }

                // Test conversion to and from an LHCbID
                {
                  const LHCb::LHCbID      lID( id );
                  const LHCb::RichSmartID nid( lID.richID() );
                  if ( nid.key() != id.key() ) {
                    cerr << "Problem with RichSmartID <-> LHCbID conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << nid << endl;
                    OK = false;
                  }
                }

                // Test conversion to/from various int types
                {
                  const std::int32_t      isHPD  = id;
                  const std::uint32_t     iuHPD  = id;
                  const std::int64_t      ilsHPD = id;
                  const std::uint64_t     iluHPD = id;
                  const LHCb::RichSmartID s_id( isHPD ), u_id( iuHPD ), sl_id( ilsHPD ), ul_id( iluHPD );
                  if ( id != s_id ) {
                    cerr << "Problem with RichSmartID <-> signed int conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << s_id << endl;
                    OK = false;
                  }
                  if ( id != u_id ) {
                    cerr << "Problem with RichSmartID <-> unsigned int conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << u_id << endl;
                    OK = false;
                  }
                  if ( id != sl_id ) {
                    cerr << "Problem with RichSmartID <-> signed long int conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << sl_id << endl;
                    OK = false;
                  }
                  if ( id != ul_id ) {
                    cerr << "Problem with RichSmartID <-> unsigned long int conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << ul_id << endl;
                    OK = false;
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // Test MaPMTs
  for ( unsigned int irich = 0; irich <= LHCb::RichSmartID::MaPMT::MaxRich; ++irich ) {
    const auto rich = (Rich::DetectorType)irich;
    for ( unsigned int ipanel = 0; ipanel <= LHCb::RichSmartID::MaPMT::MaxPanel; ++ipanel ) {
      const auto panel = (Rich::Side)ipanel;
      for ( unsigned int pdrow = 0; pdrow <= LHCb::RichSmartID::MaPMT::MaxPDNumInMod; ++pdrow ) {
        for ( unsigned int pdmod = 0; pdmod <= LHCb::RichSmartID::MaPMT::MaxPDMod; ++pdmod ) {
          for ( unsigned int pixrow = 0; pixrow <= LHCb::RichSmartID::MaPMT::MaxPixelRow; ++pixrow ) {
            for ( unsigned int pixcol = 0; pixcol <= LHCb::RichSmartID::MaPMT::MaxPixelCol; ++pixcol ) {

              // Make a test smart ID
              const LHCb::RichSmartID id( rich, panel, pdrow, pdmod, pixrow, pixcol, LHCb::RichSmartID::MaPMTID );

              // Check values
              if ( rich != id.rich() ||        //
                   panel != id.panel() ||      //
                   pdrow != id.pdNumInCol() || //
                   pdmod != id.pdCol() ||      //
                   pixrow != id.pixelRow() ||  //
                   pixcol != id.pixelCol() ) {
                cerr << "Problem with RichSmartID " << id << endl;
                OK = false;
              }

              // Test conversion to and from an LHCbID
              {
                const LHCb::LHCbID      lID( id );
                const LHCb::RichSmartID nid( lID.richID() );
                if ( nid.key() != id.key() ) {
                  cerr << "Problem with RichSmartID <-> LHCbID conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << nid << endl;
                  OK = false;
                }
              }

              // Test conversion to/from various int types
              {
                const std::int32_t      isPMT  = id;
                const std::uint32_t     iuPMT  = id;
                const std::int64_t      ilsPMT = id;
                const std::uint64_t     iluPMT = id;
                const LHCb::RichSmartID s_id( isPMT ), u_id( iuPMT ), sl_id( ilsPMT ), ul_id( iluPMT );
                if ( id != s_id ) {
                  cerr << "Problem with RichSmartID <-> signed int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << s_id << endl;
                  OK = false;
                }
                if ( id != u_id ) {
                  cerr << "Problem with RichSmartID <-> unsigned int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << u_id << endl;
                  OK = false;
                }
                if ( id != sl_id ) {
                  cerr << "Problem with RichSmartID <-> signed long int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << sl_id << endl;
                  OK = false;
                }
                if ( id != ul_id ) {
                  cerr << "Problem with RichSmartID <-> unsigned long int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << ul_id << endl;
                  OK = false;
                }

                // Test conversion to/from RichSmartID32
                {
                  const LHCb::RichSmartID32 id32( (std::uint32_t)id );
                  const LHCb::RichSmartID   new_id( id32 );
                  if ( id != new_id ) {
                    cerr << "Problem with RichSmartID <-> RichSmartID32 conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  32 bit    ID : " << id32 << endl;
                    cerr << "  Converted ID : " << new_id << endl;
                  }
                }

                // Test conversion to/from anode and EC info
                {
                  const auto anode = id.anodeIndex();
                  const auto ec    = id.elementaryCell();
                  const auto nInEC = id.pdNumInEC();
                  auto       s_id  = id;
                  s_id.setPD_EC( id.pdMod(), ec, nInEC );
                  s_id.setAnode( anode );
                  if ( id != s_id ) {
                    cerr << "Problem with RichSmartID <-> EC,Anode conversion" << endl;
                    cerr << "  Original  ID : " << id << endl;
                    cerr << "  Converted ID : " << s_id << endl;
                    OK = false;
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // Test MaPMTs with ADC time
  for ( unsigned int irich = 0; irich <= LHCb::RichSmartID::MaPMT::MaxRich; ++irich ) {
    const auto rich = (Rich::DetectorType)irich;
    for ( unsigned int ipanel = 0; ipanel <= LHCb::RichSmartID::MaPMT::MaxPanel; ++ipanel ) {
      const auto panel = (Rich::Side)ipanel;
      for ( unsigned int pdrow = 0; pdrow <= LHCb::RichSmartID::MaPMT::MaxPDNumInMod; ++pdrow ) {
        for ( unsigned int pdmod = 0; pdmod <= LHCb::RichSmartID::MaPMT::MaxPDMod; ++pdmod ) {
          LHCb::RichSmartID::ADCTimeType adcTime{10100}; // fake some ADC times.
          for ( unsigned int pixrow = 0; pixrow <= LHCb::RichSmartID::MaPMT::MaxPixelRow; ++pixrow ) {
            for ( unsigned int pixcol = 0; pixcol <= LHCb::RichSmartID::MaPMT::MaxPixelCol; ++pixcol ) {

              // Make a test smart ID
              const LHCb::RichSmartID id( rich, panel, pdrow, pdmod, pixrow, pixcol, ++adcTime );

              // Check values
              if ( rich != id.rich() ||        //
                   panel != id.panel() ||      //
                   pdrow != id.pdNumInCol() || //
                   pdmod != id.pdCol() ||      //
                   pixrow != id.pixelRow() ||  //
                   pixcol != id.pixelCol() ||  //
                   adcTime != id.adcTime() ) {
                cerr << "Problem with RichSmartID " << id << endl;
                OK = false;
              }

              // Test conversion to/from various int types (64 bit only with time).
              {
                const std::int64_t      ilsPMT = id;
                const std::uint64_t     iluPMT = id;
                const LHCb::RichSmartID sl_id( ilsPMT ), ul_id( iluPMT );
                if ( id != sl_id ) {
                  cerr << "Problem with RichSmartID <-> signed long int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << sl_id << endl;
                  OK = false;
                }
                if ( id != ul_id ) {
                  cerr << "Problem with RichSmartID <-> unsigned long int conversion" << endl;
                  cerr << "  Original  ID : " << id << endl;
                  cerr << "  Converted ID : " << ul_id << endl;
                  OK = false;
                }
              }
            }
          }
        }
      }
    }
  }

  if ( !OK ) {
    cerr << "RichSmartID tests FAILED" << endl;
  } else {
    cout << "RichSmartID tests SUCCESSFUL" << endl;
  }

  return ( OK ? 0 : 1 );
}
