/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#pragma once

#include "SIMDWrapper.h"

/**
 * Vec3 template class for handling Vector math on any base type.
 *
 * For example, if the template parameter is one of the float_v defined
 * in SIMDWrapper.h the constructor can take any combination of float, float*
 * or float_v.
 *
 * This class is intended as a Proxy class to ease the usage of SoA PoD format.
 * This class should not serve as a storage class.
 *
 * @author: Arthur Hennequin
 */
namespace LHCb::LinAlg {
  template <typename T>
  class Vec3 {
  public:
    using value_type = T;
    T x, y, z;

    Vec3() {}

    template <typename U>
    explicit Vec3( const Vec3<U>& v ) : x{v.x}, y{v.y}, z{v.z} {}

    template <typename Tx, typename Ty, typename Tz>
    constexpr Vec3( Tx x, Ty y, Tz z ) : x( x ), y( y ), z( z ) {}

    friend Vec3<T> operator+( const Vec3<T>& a, const Vec3<T>& b ) { return {a.x + b.x, a.y + b.y, a.z + b.z}; }
    friend Vec3<T> operator+( const Vec3<T>& a, const T& k ) { return {a.x + k, a.y + k, a.z + k}; }
    friend Vec3<T> operator+( const T& k, const Vec3<T>& a ) { return {k + a.x, k + a.y, k + a.z}; }

    friend Vec3<T> operator-( const Vec3<T>& a, const Vec3<T>& b ) { return {a.x - b.x, a.y - b.y, a.z - b.z}; }
    friend Vec3<T> operator-( const Vec3<T>& a, const T& k ) { return {a.x - k, a.y - k, a.z - k}; }
    friend Vec3<T> operator-( const T& k, const Vec3<T>& a ) { return {k - a.x, k - a.y, k - a.z}; }

    friend Vec3<T> operator*( const Vec3<T>& a, const Vec3<T>& b ) { return {a.x * b.x, a.y * b.y, a.z * b.z}; }
    friend Vec3<T> operator*( const Vec3<T>& a, const T& k ) { return {a.x * k, a.y * k, a.z * k}; }
    friend Vec3<T> operator*( const T& k, const Vec3<T>& a ) { return {k * a.x, k * a.y, k * a.z}; }

    friend Vec3<T> operator/( const Vec3<T>& a, const Vec3<T>& b ) { return {a.x / b.x, a.y / b.y, a.z / b.z}; }
    friend Vec3<T> operator/( const Vec3<T>& a, const T& k ) { return {a.x / k, a.y / k, a.z / k}; }
    friend Vec3<T> operator/( const T& k, const Vec3<T>& a ) { return {k / a.x, k / a.y, k / a.z}; }

    inline Vec3<T> cross( const Vec3<T>& b ) const { return {y * b.z - z * b.y, z * b.x - x * b.z, x * b.y - y * b.x}; }
    inline T       dot( const Vec3<T>& b ) const { return x * b.x + y * b.y + z * b.z; }
    inline friend T dot( const Vec3<T>& a, const Vec3<T>& b ) { return a.dot( b ); }
    inline T        mag2() const { return dot( *this ); }
    inline T        perp2() const { return x * x + y * y; }
    inline T        mag() const { return sqrt( mag2() ); }
    inline T        rho() const { return sqrt( perp2() ); }
    inline T        theta() const { return faster_atan2( rho(), z ); }
    inline T        phi() const { return faster_atan2( y, x ); }
    inline T        eta() const {
      auto zs = z / rho();
      return approx_log<T>( zs + sqrt( zs * zs + 1 ) );
    }
    inline Vec3<T> normalize() const {
      const T invMag = T{1} / sqrt( mag2() );
      return invMag * ( *this );
    }

    // Compatibility with ROOT::Math::XYZVector and friends
    T       X() const { return x; }
    T       Y() const { return y; }
    T       Z() const { return z; }
    T       Rho() const { return rho(); }
    Vec3<T> Cross( Vec3<T> const& b ) const { return cross( b ); }

    friend std::ostream& operator<<( std::ostream& s, Vec3<T> const& v ) {
      return s << '[' << v.X() << ", " << v.Y() << ", " << v.Z() << ']';
    }

    friend auto X( Vec3<T> const& v ) { return v.X(); }
    friend auto Y( Vec3<T> const& v ) { return v.Y(); }
    friend auto Z( Vec3<T> const& v ) { return v.Z(); }
  };

  template <typename Tx, typename Ty, typename Tz>
  Vec3( Tx, Ty, Tz )->Vec3<std::common_type_t<Tx, Ty, Tz>>;

  template <typename Out, typename In>
  Vec3<Out> gather( std::array<Vec3<In>, Out::size()> const& in ) {
    return std::apply(
        []( const auto&... i ) {
          return Vec3{Out{std::array{i.x.cast()...}}, Out{std::array{i.y.cast()...}}, Out{std::array{i.z.cast()...}}};
        },
        in );
  }

} // namespace LHCb::LinAlg

template <typename T>
using Vec3 = LHCb::LinAlg::Vec3<T>;

#define VEC3_SOA_ACCESSOR( name, locX, locY, locZ )                                                                    \
  template <typename T>                                                                                                \
  inline Vec3<T> name( const int key ) const {                                                                         \
    return {locX + key, locY + key, locZ + key};                                                                       \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline Vec3<T> broadcast_##name( const int key ) const {                                                             \
    return {*( locX + key ), *( locY + key ), *( locZ + key )};                                                        \
  }                                                                                                                    \
  template <typename I>                                                                                                \
  inline auto gather_##name( const I& key ) const {                                                                    \
    return LHCb::LinAlg::Vec3{gather( locX, key ), gather( locY, key ), gather( locZ, key )};                          \
  }                                                                                                                    \
  template <typename T, typename I, typename MaskT>                                                                    \
  inline Vec3<T> maskgather_##name( const I& key, const MaskT& mask, const T& src ) const {                            \
    return {maskgather( locX, key, mask, src ), maskgather( locY, key, mask, src ),                                    \
            maskgather( locZ, key, mask, src )};                                                                       \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline void store_##name( const int key, const Vec3<T>& v ) const {                                                  \
    v.x.store( locX + key );                                                                                           \
    v.y.store( locY + key );                                                                                           \
    v.z.store( locZ + key );                                                                                           \
  }                                                                                                                    \
  template <typename T, typename MaskT>                                                                                \
  inline void compressstore_##name( const int key, MaskT mask, const Vec3<T>& v ) const {                              \
    v.x.compressstore( mask, locX + key );                                                                             \
    v.y.compressstore( mask, locY + key );                                                                             \
    v.z.compressstore( mask, locZ + key );                                                                             \
  }

#define VEC3_XY_SOA_ACCESSOR( name, locX, locY, valZ )                                                                 \
  template <typename T>                                                                                                \
  inline Vec3<T> name( const int key ) const {                                                                         \
    return {locX + key, locY + key, valZ};                                                                             \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline Vec3<T> broadcast_##name( const int key ) const {                                                             \
    return {*( locX + key ), *( locY + key ), valZ};                                                                   \
  }                                                                                                                    \
  template <typename I>                                                                                                \
  inline auto gather_##name( const I& key ) const {                                                                    \
    using T = std::common_type_t<std::decay_t<decltype( gather( locX, key ) )>,                                        \
                                 std::decay_t<decltype( gather( locY, key ) )>>;                                       \
    return Vec3<T>{gather( locX, key ), gather( locY, key ), valZ};                                                    \
  }                                                                                                                    \
  template <typename T, typename I, typename MaskT>                                                                    \
  inline Vec3<T> maskgather_##name( const I& key, const MaskT& mask, const T& src ) const {                            \
    return {maskgather( locX, key, mask, src ), maskgather( locY, key, mask, src ), valZ};                             \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline void store_##name( const int key, const Vec3<T>& v ) const {                                                  \
    v.x.store( locX + key );                                                                                           \
    v.y.store( locY + key );                                                                                           \
  }                                                                                                                    \
  template <typename T, typename MaskT>                                                                                \
  inline void compressstore_##name( const int key, MaskT mask, const Vec3<T>& v ) const {                              \
    v.x.compressstore( mask, locX + key );                                                                             \
    v.y.compressstore( mask, locY + key );                                                                             \
  }

// Variadic accessors:

#define VEC3_SOA_ACCESSOR_VAR( name, locX, locY, locZ, ... )                                                           \
  template <typename T>                                                                                                \
  inline Vec3<T> name( const int key, __VA_ARGS__ ) const {                                                            \
    return {locX + key, locY + key, locZ + key};                                                                       \
  }                                                                                                                    \
  template <typename T, typename I>                                                                                    \
  inline Vec3<T> gather_##name( const I& key, __VA_ARGS__ ) const {                                                    \
    return {gather( locX, key ), gather( locY, key ), gather( locZ, key )};                                            \
  }                                                                                                                    \
  template <typename T, typename I, typename MaskT>                                                                    \
  inline Vec3<T> maskgather_##name( const I& key, const MaskT& mask, const T& src, __VA_ARGS__ ) const {               \
    return {maskgather( locX, key, mask, src ), maskgather( locY, key, mask, src ),                                    \
            maskgather( locZ, key, mask, src )};                                                                       \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline void store_##name( const int key, __VA_ARGS__, const Vec3<T>& v ) const {                                     \
    v.x.store( locX + key );                                                                                           \
    v.y.store( locY + key );                                                                                           \
    v.z.store( locZ + key );                                                                                           \
  }                                                                                                                    \
  template <typename T, typename MaskT>                                                                                \
  inline void compressstore_##name( const int key, __VA_ARGS__, MaskT mask, const Vec3<T>& v ) const {                 \
    v.x.compressstore( mask, locX + key );                                                                             \
    v.y.compressstore( mask, locY + key );                                                                             \
    v.z.compressstore( mask, locZ + key );                                                                             \
  }

#define VEC3_XY_SOA_ACCESSOR_VAR( name, locX, locY, valZ, ... )                                                        \
  template <typename T>                                                                                                \
  inline Vec3<T> name( const int key, __VA_ARGS__ ) const {                                                            \
    return {locX + key, locY + key, valZ};                                                                             \
  }                                                                                                                    \
  template <typename T, typename I>                                                                                    \
  inline Vec3<T> gather_##name( const I& key, __VA_ARGS__ ) const {                                                    \
    return {gather( locX, key ), gather( locY, key ), valZ};                                                           \
  }                                                                                                                    \
  template <typename T, typename I, typename MaskT>                                                                    \
  inline Vec3<T> maskgather_##name( const I& key, const MaskT& mask, const T& src, __VA_ARGS__ ) const {               \
    return {maskgather( locX, key, mask, src ), maskgather( locY, key, mask, src ), valZ};                             \
  }                                                                                                                    \
  template <typename T>                                                                                                \
  inline void store_##name( const int key, __VA_ARGS__, const Vec3<T>& v ) const {                                     \
    v.x.store( locX + key );                                                                                           \
    v.y.store( locY + key );                                                                                           \
  }                                                                                                                    \
  template <typename T, typename MaskT>                                                                                \
  inline void compressstore_##name( const int key, __VA_ARGS__, MaskT mask, const Vec3<T>& v ) const {                 \
    v.x.compressstore( mask, locX + key );                                                                             \
    v.y.compressstore( mask, locY + key );                                                                             \
  }
