/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#pragma once

// STL
#include <iostream>
#include <string>

// Gaudi
#include "GaudiKernel/Kernel.h"
#include "GaudiKernel/SerializeSTL.h"
#include "GaudiKernel/ToStream.h"

// RichUtils
#include "RichUtils/RichSIMDTypes.h"

// Rich Detextor
#include "RichDetectors/RichMirror.h"

namespace Rich::SIMD {

  /** @class MirrorData RichSIMDMirrorData.h
   *
   *  Utility class to cache SIMD data associated to a set of mirrors
   *
   *  @author Chris Jones
   *  @date   07/02/2018
   */
  class MirrorData final : public LHCb::SIMD::AlignedBase<LHCb::SIMD::VectorAlignment> {

  public:
    /// Basic precision (float)
    using FP = Rich::SIMD::DefaultScalarFP;
    /// SIMD version of FP
    using SIMDFP = Rich::SIMD::FP<FP>;
    /// Type for container of mirror pointers. Same size as SIMDFP.
    using Mirrors = Rich::SIMD::STDArray<const Detector::Mirror*, SIMDFP>;

  public:
    /** @class XYZ
     *  Basic representation of a (X,Y,Z) type
     */
    template <typename TYPE>
    class XYZ {
    public:
      /// Basic type alias
      using Scalar = TYPE;

    public:
      /// Coordinates
      TYPE _X{}, _Y{}, _Z{};

    public:
      // read access

      /// X coordinate
      [[nodiscard]] const TYPE& X() const noexcept { return _X; }
      /// Y coordinate
      [[nodiscard]] const TYPE& Y() const noexcept { return _Y; }
      /// Z coordinate
      [[nodiscard]] const TYPE& Z() const noexcept { return _Z; }
      /// X coordinate
      [[nodiscard]] const TYPE& x() const noexcept { return X(); }
      /// Y coordinate
      [[nodiscard]] const TYPE& y() const noexcept { return Y(); }
      /// Z coordinate
      [[nodiscard]] const TYPE& z() const noexcept { return Z(); }

    public:
      // write access

      /// X coordinate
      [[nodiscard]] TYPE& X() noexcept { return _X; }
      /// Y coordinate
      [[nodiscard]] TYPE& Y() noexcept { return _Y; }
      /// Z coordinate
      [[nodiscard]] TYPE& Z() noexcept { return _Z; }
      /// X coordinate
      [[nodiscard]] TYPE& x() noexcept { return X(); }
      /// Y coordinate
      [[nodiscard]] TYPE& y() noexcept { return Y(); }
      /// Z coordinate
      [[nodiscard]] TYPE& z() noexcept { return Z(); }

    public:
      /// send to std::ostream
      friend std::ostream& operator<<( std::ostream& s, const XYZ<TYPE>& xyz ) {
        return s << "{ " << xyz.X() << ", " << xyz.Y() << ", " << xyz.Z() << " }";
      }
    };

    /** @class Vector
     *  Basic representation of a (X,Y,Z) displacement vector
     */
    template <typename TYPE>
    class Vector : public XYZ<TYPE> {};

    /** @class Point
     *  Basic representation of a (X,Y,Z) position vector
     */
    template <typename TYPE>
    class Point : public XYZ<TYPE> {};

    // Non member operator - for points
    template <typename POINTA, typename POINTB>
    friend SIMD::Vector<FP> operator-( const POINTA& pa, const POINTB& pb ) noexcept {
      return {pa.X() - pb.X(), pa.Y() - pb.Y(), pa.Z() - pb.Z()};
    }

    /** @class Plane
     *  Basic representation of a SIMD vectorised plane.
     *  Described by the form A.x + B.y + C.z + D = 0 where (x,y,z) is
     *  any point in the plane.
     */
    template <typename TYPE>
    class Plane {
    public:
      /// Basic type alias
      using Scalar = TYPE;

    public:
      /// Plane Normal (X,Y,Z) == plane parameters (A,B,C)
      Vector<TYPE> normal;
      /// Parameter D
      TYPE D{};

    public:
      /// Normal Vector
      [[nodiscard]] const Vector<TYPE>& Normal() const noexcept { return normal; }
      /// Distance to a point
      template <typename POINT>
      decltype( auto ) Distance( const POINT& p ) const noexcept {
        return ( ( normal.X() * p.X() ) + ( normal.Y() * p.Y() ) + ( normal.Z() * p.Z() ) + D );
      }

    public:
      /// send to std::ostream
      friend std::ostream& operator<<( std::ostream& s, const Plane<TYPE>& p ) {
        return s << "{ Normal=" << p.normal << " D=" << p.D << " }";
      }
    };

  public:
    /// Default constructor
    MirrorData() = default;

  private:
    // Mirror data

    /// The mirrors for which the caches are valid
    Mirrors cache_mirrors{{}};
    /// RoC values
    alignas( LHCb::SIMD::VectorAlignment ) SIMDFP m_RoCs{};
    /// CoC point
    alignas( LHCb::SIMD::VectorAlignment ) Point<SIMDFP> m_CoCs;
    /// Normal Plane
    alignas( LHCb::SIMD::VectorAlignment ) Plane<SIMDFP> m_NormPs;

  public:
    /// Get the SIMD RoC values for the current mirrors
    [[nodiscard]] const SIMDFP& getRoCs() const noexcept { return m_RoCs; }
    /// Get the SIMD centreNormalPlane for the current mirrors
    [[nodiscard]] const Plane<SIMDFP>& getNormalPlane() const noexcept { return m_NormPs; }
    /// Get the SIM CoC for the current mirrors
    [[nodiscard]] const Point<SIMDFP>& getCoCs() const noexcept { return m_CoCs; }

  public:
    /// Update the cached information for the given mirrors
    void update( const Mirrors& mirrors ) noexcept {
      // loop over the scalar entries and update as needed
      for ( std::size_t i = 0; i < SIMDFP::Size; ++i ) {
        // the mirror pointer
        const auto m = mirrors[i];
        // If changed from before, update
        if ( m != cache_mirrors[i] ) { cache_update( m, i ); }
      }
    }

  private:
    /// Update cached information for given SIMD scalar index
    void cache_update( const Detector::Mirror* m, ///< Mirror pointer
                       const std::size_t       i  ///< SIMD scalar index to update
                       ) noexcept;

  public:
    /// send to std::ostream
    friend std::ostream& operator<<( std::ostream& s, const MirrorData& data ) {
      return s << "{ Mirrors "
               //<< data.mirrors // ToDo FixMe
               << " RoCs " << data.getRoCs() << " CoCs " << data.getCoCs() << " Plane " << data.getNormalPlane()
               << " }";
    }
  };

} // namespace Rich::SIMD
