/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once
#include "Detector/UT/ChannelID.h"
#include "Event/UTCluster.h"
#include "Kernel/UTDAQDefinitions.h"
#include "Kernel/UTStripRepresentation.h"
#include "Kernel/UTTell1ID.h"
#include "boost/container/static_vector.hpp"
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

/** @class UTTell1Board UTTell1Board.h "UTDAQ/UTTell1Board.h"
 *
 *  Class for converting from Detector::UT::ChannelID to daq channel/UTTell1ID
 *
 *  @author A. Beiter (based on code by M.Needham)
 *  @date   2018-09-04
 */

class UTTell1Board final {

  /// max # of lane in Board/RawBank
  constexpr static auto maxNumLanes = 6;

public:
  struct ExpandedChannelID {
    unsigned int sector{0};
    unsigned int module{0};
    unsigned int face{0};
    unsigned int stave{0};
    unsigned int layer{0};
    unsigned int side{0};
    unsigned int type{0};
    unsigned int uniqueSector{0};
    unsigned int chanID{0};
  };
  typedef std::pair<LHCb::Detector::UT::ChannelID, int> chanPair;

  /// constructer
  UTTell1Board( const UTTell1ID aBoard, const unsigned int stripsPerHybrid );

  /// add wafer
  void addSector( LHCb::Detector::UT::ChannelID aOfflineChan, unsigned int orientation, const std::string& serviceBox );

  /// board indentifer
  UTTell1ID boardID() const;

  /// same board id
  bool sameID( const UTTell1ID& id ) const;

  /// channel is in this board
  bool isInside( const LHCb::Detector::UT::ChannelID aOfflineChan, unsigned int& sectorIndex ) const;

  /// construct LHCb::Detector::UT::ChannelID from DAQ Channel
  chanPair DAQToOffline( const unsigned int fracStrip, const UTDAQ::version& version,
                         const UTDAQ::UTStripRepresentation aDAQChan ) const;

  /// construct LHCb::Detector::UT::ChannelID from DAQ Channel
  std::tuple<UTTell1Board::ExpandedChannelID, unsigned int, int>
  DAQToOfflineFull( const unsigned int fracStrip, const UTDAQ::version& version, unsigned int aDAQChan ) const;

  /// fill adc values offline
  void ADCToOffline( const unsigned int aDAQChan, LHCb::UTCluster::ADCVector& adcs, UTDAQ::version version,
                     const unsigned int offset, const unsigned int fracStrip ) const;

  /// construct DAQChannel from LHCb::Detector::UT::ChannelID
  unsigned int offlineToDAQ( const LHCb::Detector::UT::ChannelID aOfflineChan, const unsigned int sectorIndex,
                             double isf ) const;

  /// check channel is valid
  bool validChannel( const unsigned int daqChan ) const;

  /// vector of sectors on the board
  const boost::container::static_vector<LHCb::Detector::UT::ChannelID, maxNumLanes>& sectorIDs() const;

  /// vector of hybrid orientations
  const boost::container::static_vector<int, maxNumLanes>& orientation() const;

  /// vector of service boxes
  const std::vector<std::string>& serviceBoxes() const;

  /// service box
  std::string serviceBox( const LHCb::Detector::UT::ChannelID& chan ) const;

  /// service box
  std::string serviceBox( const UTDAQ::UTStripRepresentation& tell1Chan ) const;

  /// number of readout sectors
  unsigned int nSectors() const;

  /// Operator overloading for stringoutput
  friend std::ostream& operator<<( std::ostream& s, const UTTell1Board& obj ) { return obj.fillStream( s ); }

  // Fill the ASCII output stream
  std::ostream& fillStream( std::ostream& s ) const;

  /** print method for python Not needed in C++ */
  std::string toString() const;

private:
  /// service box
  std::string serviceBox( const unsigned int& waferIndex ) const;

  UTTell1ID                                                                   m_boardID;
  unsigned int                                                                m_nStripsPerHybrid{0};
  boost::container::static_vector<LHCb::Detector::UT::ChannelID, maxNumLanes> m_sectorsVector;
  std::vector<ExpandedChannelID>                                              m_sectorsVectorOpt;
  boost::container::static_vector<int, maxNumLanes>                           m_orientation;
  std::vector<std::string>                                                    m_serviceBoxVector;
};

inline std::string UTTell1Board::toString() const {
  std::ostringstream o;
  fillStream( o );
  return o.str();
}

inline UTTell1ID UTTell1Board::boardID() const { return m_boardID; }

inline bool UTTell1Board::sameID( const UTTell1ID& id ) const { return m_boardID == id ? true : false; }

inline const boost::container::static_vector<int, UTTell1Board::maxNumLanes>& UTTell1Board::orientation() const {
  return m_orientation;
}

inline const std::vector<std::string>& UTTell1Board::serviceBoxes() const { return m_serviceBoxVector; }

inline const boost::container::static_vector<LHCb::Detector::UT::ChannelID, UTTell1Board::maxNumLanes>&
UTTell1Board::sectorIDs() const {
  return m_sectorsVector;
}

inline std::string UTTell1Board::serviceBox( const unsigned int& waferIndex ) const {
  return ( waferIndex < m_serviceBoxVector.size() ? m_serviceBoxVector[waferIndex] : std::string( "Unknown" ) );
}

inline std::string UTTell1Board::serviceBox( const LHCb::Detector::UT::ChannelID& chan ) const {
  unsigned int waferIndex{0};
  isInside( chan, waferIndex );
  return serviceBox( waferIndex );
}

inline bool UTTell1Board::validChannel( const unsigned int daqChan ) const {
  if ( daqChan > m_nStripsPerHybrid * m_sectorsVector.size() ) return false;
  const int index = daqChan / m_nStripsPerHybrid;
  return m_sectorsVector[index].sector() != 0;
}

inline unsigned int UTTell1Board::nSectors() const { return m_sectorsVector.size(); }
